# Copyright (c) 2014 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools


from sahara.plugins import context
from sahara.plugins import exceptions as ex
from sahara.plugins import kerberos
from sahara.plugins import swift_helper
from sahara.plugins import topology_helper as t_helper
from sahara.plugins import utils
from sahara_plugin_cdh.i18n import _
from sahara_plugin_cdh.plugins.cdh.client import api_client
from sahara_plugin_cdh.plugins.cdh.client import services
from sahara_plugin_cdh.plugins.cdh import db_helper as dh
from sahara_plugin_cdh.plugins.cdh import plugin_utils
from sahara_plugin_cdh.plugins.cdh import validation


HDFS_SERVICE_TYPE = 'HDFS'
YARN_SERVICE_TYPE = 'YARN'
OOZIE_SERVICE_TYPE = 'OOZIE'
HIVE_SERVICE_TYPE = 'HIVE'
HUE_SERVICE_TYPE = 'HUE'
SPARK_SERVICE_TYPE = 'SPARK_ON_YARN'
ZOOKEEPER_SERVICE_TYPE = 'ZOOKEEPER'
HBASE_SERVICE_TYPE = 'HBASE'
FLUME_SERVICE_TYPE = 'FLUME'
SENTRY_SERVICE_TYPE = 'SENTRY'
SOLR_SERVICE_TYPE = 'SOLR'
SQOOP_SERVICE_TYPE = 'SQOOP'
KS_INDEXER_SERVICE_TYPE = 'KS_INDEXER'
IMPALA_SERVICE_TYPE = 'IMPALA'
KMS_SERVICE_TYPE = 'KMS'
KAFKA_SERVICE_TYPE = 'KAFKA'


def cloudera_cmd(f):
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        for cmd in f(*args, **kwargs):
            result = cmd.wait()
            if not result.success:
                if result.children is not None:
                    for c in result.children:
                        if not c.success:
                            raise ex.HadoopProvisionError(c.resultMessage)
                else:
                    raise ex.HadoopProvisionError(result.resultMessage)
    return wrapper


class ClouderaUtils(object):
    CM_DEFAULT_USERNAME = 'admin'
    CM_DEFAULT_PASSWD = 'admin'
    CM_API_VERSION = 8

    HDFS_SERVICE_NAME = 'hdfs01'
    YARN_SERVICE_NAME = 'yarn01'
    OOZIE_SERVICE_NAME = 'oozie01'
    HIVE_SERVICE_NAME = 'hive01'
    HUE_SERVICE_NAME = 'hue01'
    SPARK_SERVICE_NAME = 'spark_on_yarn01'
    ZOOKEEPER_SERVICE_NAME = 'zookeeper01'
    HBASE_SERVICE_NAME = 'hbase01'

    FLUME_SERVICE_NAME = 'flume01'
    SOLR_SERVICE_NAME = 'solr01'
    SQOOP_SERVICE_NAME = 'sqoop01'
    KS_INDEXER_SERVICE_NAME = 'ks_indexer01'
    IMPALA_SERVICE_NAME = 'impala01'
    SENTRY_SERVICE_NAME = 'sentry01'
    KMS_SERVICE_NAME = 'kms01'
    KAFKA_SERVICE_NAME = 'kafka01'
    NAME_SERVICE = 'nameservice01'

    def __init__(self):
        self.pu = plugin_utils.AbstractPluginUtils()
        self.validator = validation.Validator
        self.c_helper = None

    def get_api_client_by_default_password(self, cluster):
        manager_ip = self.pu.get_manager(cluster).management_ip
        return api_client.ApiResource(manager_ip,
                                      username=self.CM_DEFAULT_USERNAME,
                                      password=self.CM_DEFAULT_PASSWD,
                                      version=self.CM_API_VERSION)

    def get_api_client(self, cluster, api_version=None):
        manager_ip = self.pu.get_manager(cluster).management_ip
        cm_password = dh.get_cm_password(cluster)
        version = self.CM_API_VERSION if not api_version else api_version
        return api_client.ApiResource(manager_ip,
                                      username=self.CM_DEFAULT_USERNAME,
                                      password=cm_password,
                                      version=version)

    def update_cloudera_password(self, cluster):
        api = self.get_api_client_by_default_password(cluster)
        user = api.get_user(self.CM_DEFAULT_USERNAME)
        user.password = dh.get_cm_password(cluster)
        api.update_user(user)

    def get_cloudera_cluster(self, cluster):
        api = self.get_api_client(cluster)
        return api.get_cluster(cluster.name)

    @cloudera_cmd
    def start_cloudera_cluster(self, cluster):
        cm_cluster = self.get_cloudera_cluster(cluster)
        yield cm_cluster.start()

    @cloudera_cmd
    def stop_cloudera_cluster(self, cluster):
        cm_cluster = self.get_cloudera_cluster(cluster)
        yield cm_cluster.stop()

    def start_instances(self, cluster):
        self.start_cloudera_cluster(cluster)

    @utils.event_wrapper(
        True, step=_("Delete instances"), param=('cluster', 1))
    def delete_instances(self, cluster, instances):
        api = self.get_api_client(cluster)
        cm_cluster = self.get_cloudera_cluster(cluster)
        hosts = api.get_all_hosts(view='full')
        hostsnames_to_deleted = [i.fqdn() for i in instances]
        for host in hosts:
            if host.hostname in hostsnames_to_deleted:
                cm_cluster.remove_host(host.hostId)
                api.delete_host(host.hostId)

    @utils.event_wrapper(
        True, step=_("Decommission nodes"), param=('cluster', 1))
    def decommission_nodes(self, cluster, process,
                           decommission_roles, roles_to_delete=None):
        service = self.get_service_by_role(process, cluster)
        service.decommission(*decommission_roles).wait()
        # not all roles should be decommissioned
        if roles_to_delete:
            decommission_roles.extend(roles_to_delete)
        for role_name in decommission_roles:
            service.delete_role(role_name)

    @utils.event_wrapper(
        True, step=_("Refresh DataNodes"), param=('cluster', 1))
    def refresh_datanodes(self, cluster):
        self._refresh_nodes(cluster, 'DATANODE', self.HDFS_SERVICE_NAME)

    @utils.event_wrapper(
        True, step=_("Refresh YARNNodes"), param=('cluster', 1))
    def refresh_yarn_nodes(self, cluster):
        self._refresh_nodes(cluster, 'NODEMANAGER', self.YARN_SERVICE_NAME)

    @cloudera_cmd
    def _refresh_nodes(self, cluster, process, service_name):
        cm_cluster = self.get_cloudera_cluster(cluster)
        service = cm_cluster.get_service(service_name)
        nds = [n.name for n in service.get_roles_by_type(process)]
        for nd in nds:
            for st in service.refresh(nd):
                yield st

    @utils.event_wrapper(
        True, step=_("Restart stale services"), param=('cluster', 1))
    @cloudera_cmd
    def restart_stale_services(self, cluster):
        cm_cluster = self.get_cloudera_cluster(cluster)
        yield cm_cluster.restart(
            restart_only_stale_services=True,
            redeploy_client_configuration=True)

    @utils.event_wrapper(True, step=_("Deploy configs"), param=('cluster', 1))
    @cloudera_cmd
    def deploy_configs(self, cluster):
        cm_cluster = self.get_cloudera_cluster(cluster)
        yield cm_cluster.deploy_client_config()

    def update_configs(self, instances):
        # instances non-empty
        utils.add_provisioning_step(
            instances[0].cluster_id, _("Update configs"), len(instances))
        with context.PluginsThreadGroup() as tg:
            for instance in instances:
                tg.spawn("update-configs-%s" % instance.instance_name,
                         self._update_configs, instance)
                context.sleep(1)

    @utils.event_wrapper(True)
    @cloudera_cmd
    def _update_configs(self, instance):
        for process in instance.node_group.node_processes:
            process = self.pu.convert_role_showname(process)
            service = self.get_service_by_role(process, instance=instance)
            yield service.deploy_client_config(self.pu.get_role_name(instance,
                                                                     process))

    def get_mgmt_service(self, cluster):
        api = self.get_api_client(cluster)
        cm = api.get_cloudera_manager()
        mgmt_service = cm.get_service()
        return mgmt_service

    @cloudera_cmd
    def restart_mgmt_service(self, cluster):
        service = self.get_mgmt_service(cluster)
        yield service.restart()

    @cloudera_cmd
    def start_service(self, service):
        yield service.start()

    @cloudera_cmd
    def stop_service(self, service):
        yield service.stop()

    @cloudera_cmd
    def start_roles(self, service, *role_names):
        for role in service.start_roles(*role_names):
            yield role

    @utils.event_wrapper(
        True, step=_("Create mgmt service"), param=('cluster', 1))
    def create_mgmt_service(self, cluster):
        api = self.get_api_client(cluster)
        cm = api.get_cloudera_manager()

        setup_info = services.ApiServiceSetupInfo()
        manager = self.pu.get_manager(cluster)
        hostname = manager.fqdn()
        processes = ['SERVICEMONITOR', 'HOSTMONITOR',
                     'EVENTSERVER', 'ALERTPUBLISHER']
        for proc in processes:
            setup_info.add_role_info(self.pu.get_role_name(manager, proc),
                                     proc, hostname)

        cm.create_mgmt_service(setup_info)
        cm.hosts_start_roles([hostname])

    def get_service_by_role(self, role, cluster=None, instance=None):
        if cluster:
            cm_cluster = self.get_cloudera_cluster(cluster)
        elif instance:
            cm_cluster = self.get_cloudera_cluster(instance.cluster)
        else:
            raise ValueError(_("'cluster' or 'instance' argument missed"))

        if role in ['NAMENODE', 'DATANODE', 'SECONDARYNAMENODE',
                    'HDFS_GATEWAY']:
            return cm_cluster.get_service(self.HDFS_SERVICE_NAME)
        elif role in ['RESOURCEMANAGER', 'NODEMANAGER', 'JOBHISTORY',
                      'YARN_GATEWAY']:
            return cm_cluster.get_service(self.YARN_SERVICE_NAME)
        elif role in ['OOZIE_SERVER']:
            return cm_cluster.get_service(self.OOZIE_SERVICE_NAME)
        elif role in ['HIVESERVER2', 'HIVEMETASTORE', 'WEBHCAT']:
            return cm_cluster.get_service(self.HIVE_SERVICE_NAME)
        elif role in ['HUE_SERVER']:
            return cm_cluster.get_service(self.HUE_SERVICE_NAME)
        elif role in ['SPARK_YARN_HISTORY_SERVER']:
            return cm_cluster.get_service(self.SPARK_SERVICE_NAME)
        elif role in ['SERVER']:
            return cm_cluster.get_service(self.ZOOKEEPER_SERVICE_NAME)
        elif role in ['MASTER', 'REGIONSERVER']:
            return cm_cluster.get_service(self.HBASE_SERVICE_NAME)
        elif role in ['AGENT']:
            return cm_cluster.get_service(self.FLUME_SERVICE_NAME)
        elif role in ['SENTRY_SERVER']:
            return cm_cluster.get_service(self.SENTRY_SERVICE_NAME)
        elif role in ['SQOOP_SERVER']:
            return cm_cluster.get_service(self.SQOOP_SERVICE_NAME)
        elif role in ['SOLR_SERVER']:
            return cm_cluster.get_service(self.SOLR_SERVICE_NAME)
        elif role in ['HBASE_INDEXER']:
            return cm_cluster.get_service(self.KS_INDEXER_SERVICE_NAME)
        elif role in ['CATALOGSERVER', 'STATESTORE', 'IMPALAD', 'LLAMA']:
            return cm_cluster.get_service(self.IMPALA_SERVICE_NAME)
        elif role in ['KMS']:
            return cm_cluster.get_service(self.KMS_SERVICE_NAME)
        elif role in ['JOURNALNODE']:
            return cm_cluster.get_service(self.HDFS_SERVICE_NAME)
        elif role in ['YARN_STANDBYRM']:
            return cm_cluster.get_service(self.YARN_SERVICE_NAME)
        elif role in ['KAFKA_BROKER']:
            return cm_cluster.get_service(self.KAFKA_SERVICE_NAME)
        else:
            raise ValueError(
                _("Process %(process)s is not supported by CDH plugin") %
                {'process': role})

    @utils.event_wrapper(
        True, step=_("First run cluster"), param=('cluster', 1))
    @cloudera_cmd
    def first_run(self, cluster):
        cm_cluster = self.get_cloudera_cluster(cluster)
        yield cm_cluster.first_run()

    @utils.event_wrapper(
        True, step=_("Create services"), param=('cluster', 1))
    def create_services(self, cluster):
        api = self.get_api_client(cluster)
        cm_cluster = api.create_cluster(cluster.name,
                                        fullVersion=cluster.hadoop_version)

        if len(self.pu.get_zookeepers(cluster)) > 0:
            cm_cluster.create_service(self.ZOOKEEPER_SERVICE_NAME,
                                      ZOOKEEPER_SERVICE_TYPE)
        cm_cluster.create_service(self.HDFS_SERVICE_NAME, HDFS_SERVICE_TYPE)
        cm_cluster.create_service(self.YARN_SERVICE_NAME, YARN_SERVICE_TYPE)
        cm_cluster.create_service(self.OOZIE_SERVICE_NAME, OOZIE_SERVICE_TYPE)
        if self.pu.get_hive_metastore(cluster):
            cm_cluster.create_service(self.HIVE_SERVICE_NAME,
                                      HIVE_SERVICE_TYPE)
        if self.pu.get_hue(cluster):
            cm_cluster.create_service(self.HUE_SERVICE_NAME, HUE_SERVICE_TYPE)
        if self.pu.get_spark_historyserver(cluster):
            cm_cluster.create_service(self.SPARK_SERVICE_NAME,
                                      SPARK_SERVICE_TYPE)
        if self.pu.get_hbase_master(cluster):
            cm_cluster.create_service(self.HBASE_SERVICE_NAME,
                                      HBASE_SERVICE_TYPE)
        if len(self.pu.get_flumes(cluster)) > 0:
            cm_cluster.create_service(self.FLUME_SERVICE_NAME,
                                      FLUME_SERVICE_TYPE)
        if self.pu.get_sentry(cluster):
            cm_cluster.create_service(self.SENTRY_SERVICE_NAME,
                                      SENTRY_SERVICE_TYPE)
        if len(self.pu.get_solrs(cluster)) > 0:
            cm_cluster.create_service(self.SOLR_SERVICE_NAME,
                                      SOLR_SERVICE_TYPE)
        if self.pu.get_sqoop(cluster):
            cm_cluster.create_service(self.SQOOP_SERVICE_NAME,
                                      SQOOP_SERVICE_TYPE)
        if len(self.pu.get_hbase_indexers(cluster)) > 0:
            cm_cluster.create_service(self.KS_INDEXER_SERVICE_NAME,
                                      KS_INDEXER_SERVICE_TYPE)
        if self.pu.get_catalogserver(cluster):
            cm_cluster.create_service(self.IMPALA_SERVICE_NAME,
                                      IMPALA_SERVICE_TYPE)
        if self.pu.get_kms(cluster):
            cm_cluster.create_service(self.KMS_SERVICE_NAME,
                                      KMS_SERVICE_TYPE)
        if len(self.pu.get_kafka_brokers(cluster)) > 0:
            cm_cluster.create_service(self.KAFKA_SERVICE_NAME,
                                      KAFKA_SERVICE_TYPE)

    def _agents_connected(self, instances, api):
        hostnames = [i.fqdn() for i in instances]
        hostnames_to_manager = [h.hostname for h in
                                api.get_all_hosts('full')]
        for hostname in hostnames:
            if hostname not in hostnames_to_manager:
                return False
        return True

    @utils.event_wrapper(True, step=_("Await agents"), param=('cluster', 1))
    def _await_agents(self, cluster, instances, timeout_config):
        api = self.get_api_client(instances[0].cluster)
        utils.plugin_option_poll(
            cluster, self._agents_connected, timeout_config,
            _("Await Cloudera agents"), 5, {
                'instances': instances, 'api': api})

    def await_agents(self, cluster, instances):
        self._await_agents(cluster, instances,
                           self.c_helper.AWAIT_AGENTS_TIMEOUT)

    @utils.event_wrapper(
        True, step=_("Configure services"), param=('cluster', 1))
    def configure_services(self, cluster):
        cm_cluster = self.get_cloudera_cluster(cluster)

        if len(self.pu.get_zookeepers(cluster)) > 0:
            zookeeper = cm_cluster.get_service(self.ZOOKEEPER_SERVICE_NAME)
            zookeeper.update_config(self._get_configs(ZOOKEEPER_SERVICE_TYPE,
                                                      cluster=cluster))

        hdfs = cm_cluster.get_service(self.HDFS_SERVICE_NAME)
        hdfs.update_config(self._get_configs(HDFS_SERVICE_TYPE,
                                             cluster=cluster))

        yarn = cm_cluster.get_service(self.YARN_SERVICE_NAME)
        yarn.update_config(self._get_configs(YARN_SERVICE_TYPE,
                                             cluster=cluster))

        oozie = cm_cluster.get_service(self.OOZIE_SERVICE_NAME)
        oozie.update_config(self._get_configs(OOZIE_SERVICE_TYPE,
                                              cluster=cluster))

        if self.pu.get_hive_metastore(cluster):
            hive = cm_cluster.get_service(self.HIVE_SERVICE_NAME)
            hive.update_config(self._get_configs(HIVE_SERVICE_TYPE,
                                                 cluster=cluster))

        if self.pu.get_hue(cluster):
            hue = cm_cluster.get_service(self.HUE_SERVICE_NAME)
            hue.update_config(self._get_configs(HUE_SERVICE_TYPE,
                                                cluster=cluster))

        if self.pu.get_spark_historyserver(cluster):
            spark = cm_cluster.get_service(self.SPARK_SERVICE_NAME)
            spark.update_config(self._get_configs(SPARK_SERVICE_TYPE,
                                                  cluster=cluster))

        if self.pu.get_hbase_master(cluster):
            hbase = cm_cluster.get_service(self.HBASE_SERVICE_NAME)
            hbase.update_config(self._get_configs(HBASE_SERVICE_TYPE,
                                                  cluster=cluster))

        if len(self.pu.get_flumes(cluster)) > 0:
            flume = cm_cluster.get_service(self.FLUME_SERVICE_NAME)
            flume.update_config(self._get_configs(FLUME_SERVICE_TYPE,
                                                  cluster=cluster))

        if self.pu.get_sentry(cluster):
            sentry = cm_cluster.get_service(self.SENTRY_SERVICE_NAME)
            sentry.update_config(self._get_configs(SENTRY_SERVICE_TYPE,
                                                   cluster=cluster))

        if len(self.pu.get_solrs(cluster)) > 0:
            solr = cm_cluster.get_service(self.SOLR_SERVICE_NAME)
            solr.update_config(self._get_configs(SOLR_SERVICE_TYPE,
                                                 cluster=cluster))

        if self.pu.get_sqoop(cluster):
            sqoop = cm_cluster.get_service(self.SQOOP_SERVICE_NAME)
            sqoop.update_config(self._get_configs(SQOOP_SERVICE_TYPE,
                                                  cluster=cluster))

        if len(self.pu.get_hbase_indexers(cluster)) > 0:
            ks_indexer = cm_cluster.get_service(self.KS_INDEXER_SERVICE_NAME)
            ks_indexer.update_config(
                self._get_configs(KS_INDEXER_SERVICE_TYPE, cluster=cluster))

        if self.pu.get_catalogserver(cluster):
            impala = cm_cluster.get_service(self.IMPALA_SERVICE_NAME)
            impala.update_config(self._get_configs(IMPALA_SERVICE_TYPE,
                                                   cluster=cluster))

        if self.pu.get_kms(cluster):
            kms = cm_cluster.get_service(self.KMS_SERVICE_NAME)
            kms.update_config(self._get_configs(KMS_SERVICE_TYPE,
                                                cluster=cluster))
        if len(self.pu.get_kafka_brokers(cluster)) > 0:
            kafka = cm_cluster.get_service(self.KAFKA_SERVICE_NAME)
            kafka.update_config(self._get_configs(KAFKA_SERVICE_TYPE,
                                                  cluster=cluster))

    def configure_instances(self, instances, cluster=None):
        # instances non-empty
        utils.add_provisioning_step(
            instances[0].cluster_id, _("Configure instances"), len(instances))
        for inst in instances:
            self.configure_instance(inst, cluster)

    def get_roles_list(self, node_processes):
        current = set(node_processes)
        extra_roles = {
            'YARN_GATEWAY': ["YARN_NODEMANAGER"],
            'HDFS_GATEWAY': ['HDFS_NAMENODE', 'HDFS_DATANODE',
                             "HDFS_SECONDARYNAMENODE"]
        }
        for extra_role in extra_roles.keys():
            valid_processes = extra_roles[extra_role]
            for valid in valid_processes:
                if valid in current:
                    current.add(extra_role)
                    break
        return list(current)

    def get_role_type(self, process):
        mapper = {
            'YARN_GATEWAY': 'GATEWAY',
            'HDFS_GATEWAY': 'GATEWAY',
        }
        return mapper.get(process, process)

    @utils.event_wrapper(True)
    def configure_instance(self, instance, cluster=None):
        roles_list = self.get_roles_list(instance.node_group.node_processes)
        for role in roles_list:
            self._add_role(instance, role, cluster)

    def _add_role(self, instance, process, cluster):
        if process in ['CLOUDERA_MANAGER', 'HDFS_JOURNALNODE',
                       'YARN_STANDBYRM']:
            return

        process = self.pu.convert_role_showname(process)
        service = self.get_service_by_role(process, instance=instance)
        role_type = self.get_role_type(process)
        role = service.create_role(self.pu.get_role_name(instance, process),
                                   role_type, instance.fqdn())
        role.update_config(self._get_configs(process, cluster,
                                             instance=instance))

    @cloudera_cmd
    def restart_service(self, process, instance):
        service = self.get_service_by_role(process, instance=instance)
        yield service.restart()

    def update_role_config(self, instance, process):
        process = self.pu.convert_role_showname(process)
        service = self.get_service_by_role(process, instance=instance)
        api = self.get_api_client(instance.cluster)
        hosts = api.get_all_hosts(view='full')
        ihost_id = None
        for host in hosts:
            if instance.fqdn() == host.hostname:
                ihost_id = host.hostId
                break
        role_type = self.get_role_type(process)
        roles = service.get_roles_by_type(role_type)
        for role in roles:
            if role.hostRef.hostId == ihost_id:
                role.update_config(
                    self._get_configs(role_type, instance=instance))
        self.restart_service(process, instance)

    @cloudera_cmd
    def import_admin_credentials(self, cm, username, password):
        yield cm.import_admin_credentials(username, password)

    @cloudera_cmd
    def configure_for_kerberos(self, cluster):
        api = self.get_api_client(cluster, api_version=11)
        cluster = api.get_cluster(cluster.name)
        yield cluster.configure_for_kerberos()

    def push_kerberos_configs(self, cluster):
        manager = self.pu.get_manager(cluster)
        kdc_host = kerberos.get_kdc_host(cluster, manager)
        security_realm = kerberos.get_realm_name(cluster)
        username = "%s@%s" % (kerberos.get_admin_principal(cluster),
                              kerberos.get_realm_name(cluster))
        password = kerberos.get_server_password(cluster)

        api = self.get_api_client(cluster)
        cm = api.get_cloudera_manager()
        cm.update_config({'SECURITY_REALM': security_realm,
                          'KDC_HOST': kdc_host})

        self.import_admin_credentials(cm, username, password)
        self.configure_for_kerberos(cluster)
        self.deploy_configs(cluster)

    def configure_rack_awareness(self, cluster):
        if t_helper.is_data_locality_enabled():
            self._configure_rack_awareness(cluster)

    @utils.event_wrapper(
        True, step=_("Configure rack awareness"), param=('cluster', 1))
    def _configure_rack_awareness(self, cluster):
        api = self.get_api_client(cluster)
        topology = t_helper.generate_topology_map(
            cluster, is_node_awareness=False)
        for host in api.get_all_hosts():
            host.rackId = topology[host.ipAddress]
            host.put_host()

    def full_cluster_stop(self, cluster):
        self.stop_cloudera_cluster(cluster)
        mgmt = self.get_mgmt_service(cluster)
        self.stop_service(mgmt)

    def full_cluster_start(self, cluster):
        self.start_cloudera_cluster(cluster)
        mgmt = self.get_mgmt_service(cluster)
        self.start_service(mgmt)

    def get_cloudera_manager_info(self, cluster):
        mng = self.pu.get_manager(cluster)
        info = {
            'Cloudera Manager': {
                'Web UI': 'http://%s:7180' % mng.get_ip_or_dns_name(),
                'Username': 'admin',
                'Password': dh.get_cm_password(cluster)
            }
        }
        return info

    @utils.event_wrapper(
        True, step=_("Enable NameNode HA"), param=('cluster', 1))
    @cloudera_cmd
    def enable_namenode_ha(self, cluster):
        standby_nn = self.pu.get_secondarynamenode(cluster)
        standby_nn_host_name = standby_nn.fqdn()
        jns = self.pu.get_jns(cluster)
        jn_list = []
        for index, jn in enumerate(jns):
            jn_host_name = jn.fqdn()
            jn_list.append({'jnHostId': jn_host_name,
                            'jnName': 'JN%i' % index,
                            'jnEditsDir': '/dfs/jn'
                            })
        cm_cluster = self.get_cloudera_cluster(cluster)
        hdfs = cm_cluster.get_service(self.HDFS_SERVICE_NAME)
        nn = hdfs.get_roles_by_type('NAMENODE')[0]

        yield hdfs.enable_nn_ha(active_name=nn.name,
                                standby_host_id=standby_nn_host_name,
                                nameservice=self.NAME_SERVICE, jns=jn_list
                                )

    @utils.event_wrapper(
        True, step=_("Enable ResourceManager HA"), param=('cluster', 1))
    @cloudera_cmd
    def enable_resourcemanager_ha(self, cluster):
        new_rm = self.pu.get_stdb_rm(cluster)
        new_rm_host_name = new_rm.fqdn()
        cm_cluster = self.get_cloudera_cluster(cluster)
        yarn = cm_cluster.get_service(self.YARN_SERVICE_NAME)
        yield yarn.enable_rm_ha(new_rm_host_id=new_rm_host_name)

    def _load_version_specific_instance_configs(self, instance, default_conf):
        pass

    def _get_configs(self, service, cluster=None, instance=None):
        def get_hadoop_dirs(mount_points, suffix):
            return ','.join([x + suffix for x in mount_points])

        all_confs = {}
        if cluster:
            zk_count = self.validator.get_inst_count(cluster,
                                                     'ZOOKEEPER_SERVER')
            hbm_count = self.validator.get_inst_count(cluster, 'HBASE_MASTER')
            snt_count = self.validator.get_inst_count(cluster,
                                                      'SENTRY_SERVER')
            ks_count =\
                self.validator.get_inst_count(cluster,
                                              'KEY_VALUE_STORE_INDEXER')
            kms_count = self.validator.get_inst_count(cluster, 'KMS')
            imp_count =\
                self.validator.get_inst_count(cluster,
                                              'IMPALA_CATALOGSERVER')
            hive_count = self.validator.get_inst_count(cluster,
                                                       'HIVE_METASTORE')
            slr_count = self.validator.get_inst_count(cluster, 'SOLR_SERVER')
            sqp_count = self.validator.get_inst_count(cluster, 'SQOOP_SERVER')
            core_site_safety_valve = ''
            if self.pu.c_helper.is_swift_enabled(cluster):
                configs = swift_helper.get_swift_configs()
                confs = {c['name']: c['value'] for c in configs}
                core_site_safety_valve = utils.create_elements_xml(confs)
            all_confs = {
                'HDFS': {
                    'zookeeper_service':
                        self.ZOOKEEPER_SERVICE_NAME if zk_count else '',
                    'dfs_block_local_path_access_user':
                        'impala' if imp_count else '',
                    'kms_service': self.KMS_SERVICE_NAME if kms_count else '',
                    'core_site_safety_valve': core_site_safety_valve
                },
                'HIVE': {
                    'mapreduce_yarn_service': self.YARN_SERVICE_NAME,
                    'sentry_service':
                        self.SENTRY_SERVICE_NAME if snt_count else '',
                    'zookeeper_service':
                        self.ZOOKEEPER_SERVICE_NAME if zk_count else ''
                },
                'OOZIE': {
                    'mapreduce_yarn_service': self.YARN_SERVICE_NAME,
                    'hive_service':
                        self.HIVE_SERVICE_NAME if hive_count else '',
                    'zookeeper_service':
                        self.ZOOKEEPER_SERVICE_NAME if zk_count else ''
                },
                'YARN': {
                    'hdfs_service': self.HDFS_SERVICE_NAME,
                    'zookeeper_service':
                        self.ZOOKEEPER_SERVICE_NAME if zk_count else ''
                },
                'HUE': {
                    'hive_service': self.HIVE_SERVICE_NAME,
                    'oozie_service': self.OOZIE_SERVICE_NAME,
                    'sentry_service':
                        self.SENTRY_SERVICE_NAME if snt_count else '',
                    'solr_service':
                        self.SOLR_SERVICE_NAME if slr_count else '',
                    'zookeeper_service':
                        self.ZOOKEEPER_SERVICE_NAME if zk_count else '',
                    'hbase_service':
                        self.HBASE_SERVICE_NAME if hbm_count else '',
                    'impala_service':
                        self.IMPALA_SERVICE_NAME if imp_count else '',
                    'sqoop_service':
                        self.SQOOP_SERVICE_NAME if sqp_count else ''
                },
                'SPARK_ON_YARN': {
                    'yarn_service': self.YARN_SERVICE_NAME
                },
                'HBASE': {
                    'hdfs_service': self.HDFS_SERVICE_NAME,
                    'zookeeper_service': self.ZOOKEEPER_SERVICE_NAME,
                    'hbase_enable_indexing': 'true' if ks_count else 'false',
                    'hbase_enable_replication':
                        'true' if ks_count else 'false'
                },
                'FLUME': {
                    'hdfs_service': self.HDFS_SERVICE_NAME,
                    'solr_service':
                        self.SOLR_SERVICE_NAME if slr_count else '',
                    'hbase_service':
                        self.HBASE_SERVICE_NAME if hbm_count else ''
                },
                'SENTRY': {
                    'hdfs_service': self.HDFS_SERVICE_NAME,
                    'sentry_server_config_safety_valve': (
                        self.c_helper.SENTRY_IMPALA_CLIENT_SAFETY_VALVE
                        if imp_count else '')
                },
                'SOLR': {
                    'hdfs_service': self.HDFS_SERVICE_NAME,
                    'zookeeper_service': self.ZOOKEEPER_SERVICE_NAME
                },
                'SQOOP': {
                    'mapreduce_yarn_service': self.YARN_SERVICE_NAME
                },
                'KS_INDEXER': {
                    'hbase_service': self.HBASE_SERVICE_NAME,
                    'solr_service': self.SOLR_SERVICE_NAME
                },
                'IMPALA': {
                    'hdfs_service': self.HDFS_SERVICE_NAME,
                    'hbase_service':
                        self.HBASE_SERVICE_NAME if hbm_count else '',
                    'hive_service': self.HIVE_SERVICE_NAME,
                    'sentry_service':
                        self.SENTRY_SERVICE_NAME if snt_count else '',
                    'zookeeper_service':
                        self.ZOOKEEPER_SERVICE_NAME if zk_count else ''
                }
            }
            hive_confs = {
                'HIVE': {
                    'hive_metastore_database_type': 'postgresql',
                    'hive_metastore_database_host':
                        self.pu.get_manager(cluster).internal_ip,
                    'hive_metastore_database_port': '7432',
                    'hive_metastore_database_password':
                        dh.get_hive_db_password(cluster)
                }
            }
            hue_confs = {
                'HUE': {
                    'hue_webhdfs': self.pu.get_role_name(
                        self.pu.get_namenode(cluster), 'NAMENODE')
                }
            }
            sentry_confs = {
                'SENTRY': {
                    'sentry_server_database_type': 'postgresql',
                    'sentry_server_database_host':
                        self.pu.get_manager(cluster).internal_ip,
                    'sentry_server_database_port': '7432',
                    'sentry_server_database_password':
                        dh.get_sentry_db_password(cluster)
                }
            }
            kafka_confs = {
                'KAFKA': {
                    'zookeeper_service': self.ZOOKEEPER_SERVICE_NAME
                }
            }
            all_confs = utils.merge_configs(all_confs, hue_confs)
            all_confs = utils.merge_configs(all_confs, hive_confs)
            all_confs = utils.merge_configs(all_confs, sentry_confs)
            all_confs = utils.merge_configs(all_confs, kafka_confs)
            all_confs = utils.merge_configs(all_confs, cluster.cluster_configs)

        if instance:
            snt_count = self.validator.get_inst_count(instance.cluster,
                                                      'SENTRY_SERVER')
            paths = instance.storage_paths()

            instance_default_confs = {
                'NAMENODE': {
                    'dfs_name_dir_list': get_hadoop_dirs(paths, '/fs/nn')
                },
                'SECONDARYNAMENODE': {
                    'fs_checkpoint_dir_list':
                        get_hadoop_dirs(paths, '/fs/snn')
                },
                'DATANODE': {
                    'dfs_data_dir_list': get_hadoop_dirs(paths, '/fs/dn'),
                    'dfs_datanode_data_dir_perm': 755,
                    'dfs_datanode_handler_count': 30
                },
                'NODEMANAGER': {
                    'yarn_nodemanager_local_dirs':
                        get_hadoop_dirs(paths, '/yarn/local'),
                    'container_executor_allowed_system_users':
                        "nobody,impala,hive,llama,hdfs,yarn,mapred,"
                        "spark,oozie",
                    "container_executor_banned_users": "bin"
                },
                'SERVER': {
                    'maxSessionTimeout': 60000
                },
                'HIVESERVER2': {
                    'hiveserver2_enable_impersonation':
                        'false' if snt_count else 'true',
                    'hive_hs2_config_safety_valve': (
                        self.c_helper.HIVE_SERVER2_SENTRY_SAFETY_VALVE
                        if snt_count else '')
                },
                'HIVEMETASTORE': {
                    'hive_metastore_config_safety_valve': (
                        self.c_helper.HIVE_METASTORE_SENTRY_SAFETY_VALVE
                        if snt_count else '')
                }
            }

            self._load_version_specific_instance_configs(
                instance, instance_default_confs)

            ng_user_confs = self.pu.convert_process_configs(
                instance.node_group.node_configs)
            all_confs = utils.merge_configs(all_confs, ng_user_confs)
            all_confs = utils.merge_configs(all_confs, instance_default_confs)

        return all_confs.get(service, {})



# Copyright (c) 2015 Intel Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

from sahara.plugins import base as pb
from sahara_plugin_cdh.plugins.cdh.v5_11_0 import edp_engine
from sahara_plugin_cdh.plugins.cdh.v5_11_0 import versionhandler
from sahara_plugin_cdh.tests.unit import base


class VersionHandlerTestCase(base.SaharaTestCase):

    plugin_path = "sahara_plugin_cdh.plugins.cdh.v5_11_0."
    cloudera_utils_path = plugin_path + "cloudera_utils.ClouderaUtilsV5110."
    plugin_utils_path = plugin_path + "plugin_utils.PluginUtilsV5110."

    def setUp(self):
        super(VersionHandlerTestCase, self).setUp()
        self.vh = versionhandler.VersionHandler()
        self.override_config('plugins', ['cdh'])
        pb.setup_plugins()

    def test_get_node_processes(self):
        processes = self.vh.get_node_processes()
        for k, v in processes.items():
            for p in v:
                self.assertIsInstance(p, str)

    @mock.patch("sahara.plugins.conductor.cluster_update")
    @mock.patch("sahara.plugins.context.ctx")
    @mock.patch(plugin_path + "deploy.configure_cluster")
    @mock.patch(cloudera_utils_path + "get_cloudera_manager_info",
                return_value={"fake_cm_info": "fake"})
    def test_config_cluster(self, get_cm_info, configure_cluster,
                            ctx, cluster_update):
        cluster = mock.Mock()
        self.vh.configure_cluster(cluster)
        configure_cluster.assert_called_once_with(cluster)
        cluster_update.assert_called_once_with(
            ctx(), cluster,
            {'info': {"fake_cm_info": "fake"}})

    @mock.patch(plugin_path + "deploy.start_cluster")
    def test_start_cluster(self, start_cluster):
        cluster = mock.Mock()
        self.vh._set_cluster_info = mock.Mock()
        self.vh.start_cluster(cluster)
        start_cluster.assert_called_once_with(cluster)
        self.vh._set_cluster_info.assert_called_once_with(cluster)

    @mock.patch(plugin_path + "deploy.decommission_cluster")
    def test_decommission_nodes(self, decommission_cluster):
        cluster = mock.Mock()
        instances = mock.Mock()
        self.vh.decommission_nodes(cluster, instances)
        decommission_cluster.assert_called_once_with(cluster,
                                                     instances)

    @mock.patch(plugin_path + "deploy.scale_cluster")
    def test_scale_cluster(self, scale_cluster):
        cluster = mock.Mock()
        instances = mock.Mock()
        self.vh.scale_cluster(cluster, instances)
        scale_cluster.assert_called_once_with(cluster, instances)

    @mock.patch("sahara.plugins.conductor.cluster_update")
    @mock.patch("sahara.plugins.context.ctx")
    @mock.patch(cloudera_utils_path + "get_cloudera_manager_info",
                return_value={})
    @mock.patch(plugin_utils_path + "get_hue")
    def test_set_cluster_info(self, get_hue, get_cloudera_manager_info,
                              ctx, cluster_update):
        hue = mock.Mock()
        hue.get_ip_or_dns_name.return_value = "1.2.3.4"
        get_hue.return_value = hue
        cluster = mock.Mock()
        self.vh._set_cluster_info(cluster)
        info = {'info': {'Hue Dashboard': {'Web UI': 'http://1.2.3.4:8888'}}}
        cluster_update.assert_called_once_with(ctx(), cluster, info)

    @mock.patch("sahara.plugins.utils.get_instance")
    @mock.patch("sahara.plugins.utils.get_config_value_or_default")
    @mock.patch("sahara.plugins.edp.get_plugin")
    def test_get_edp_engine(self, get_plugin, get_config_value_or_default,
                            get_instance):
        cluster = mock.Mock()
        job_type = 'Java'
        ret = self.vh.get_edp_engine(cluster, job_type)
        self.assertIsInstance(ret, edp_engine.EdpOozieEngine)

        job_type = 'Spark'
        ret = self.vh.get_edp_engine(cluster, job_type)
        self.assertIsInstance(ret, edp_engine.EdpSparkEngine)

        job_type = 'unsupported'
        ret = self.vh.get_edp_engine(cluster, job_type)
        self.assertIsNone(ret)

    def test_get_edp_job_types(self):
        ret = self.vh.get_edp_job_types()
        expect = edp_engine.EdpOozieEngine.get_supported_job_types() + \
            edp_engine.EdpSparkEngine.get_supported_job_types()
        self.assertEqual(expect, ret)

    @mock.patch(plugin_path +
                "edp_engine.EdpOozieEngine.get_possible_job_config",
                return_value={'job_config': {}})
    def test_edp_config_hints(self, get_possible_job_config):
        job_type = mock.Mock()
        ret = self.vh.get_edp_config_hints(job_type)
        get_possible_job_config.assert_called_once_with(job_type)
        self.assertEqual(ret, {'job_config': {}})

    @mock.patch(plugin_path + "deploy.get_open_ports", return_value=[1234])
    def test_get_open_ports(self, get_open_ports):
        node_group = mock.Mock()
        ret = self.vh.get_open_ports(node_group)
        get_open_ports.assert_called_once_with(node_group)
        self.assertEqual(ret, [1234])

    @mock.patch(plugin_utils_path + "recommend_configs")
    def test_recommend_configs(self, recommend_configs):
        cluster = mock.Mock()
        scaling = mock.Mock()
        self.vh.get_plugin_configs = mock.Mock()
        self.vh.recommend_configs(cluster, scaling)
        recommend_configs.assert_called_once_with(cluster,
                                                  self.vh.get_plugin_configs(),
                                                  scaling)



# Copyright (c) 2014 Intel Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# The contents of this file are mainly copied from cm_api sources,
# released by Cloudera. Codes not used by Sahara CDH plugin are removed.
# You can find the original codes at
#
#     https://github.com/cloudera/cm_api/tree/master/python/src/cm_api
#
# To satisfy the pep8 and python3 tests, we did some changes to the codes.
# We also change some importings to use Sahara inherited classes.

from sahara_plugin_cdh.i18n import _
from sahara_plugin_cdh.plugins.cdh.client import services
from sahara_plugin_cdh.plugins.cdh.client import types
from sahara_plugin_cdh.plugins.cdh import exceptions as ex

CLUSTERS_PATH = "/clusters"


def create_cluster(resource_root, name, version=None, fullVersion=None):
    """Create a cluster

    :param resource_root: The root Resource object.
    :param name: Cluster name
    :param version: Cluster CDH major version (eg: "CDH4")
                    - The CDH minor version will be assumed to be the
                      latest released version for CDH4, or 5.0 for CDH5.
    :param fullVersion: Cluster's full CDH version. (eg: "5.1.1")
                        - If specified, 'version' will be ignored.
                        - Since: v6
    :return: An ApiCluster object
    """
    if version is None and fullVersion is None:
        raise ex.CMApiVersionError(
            _("Either 'version' or 'fullVersion' must be specified"))
    if fullVersion is not None:
        api_version = 6
        version = None
    else:
        api_version = 1

    apicluster = ApiCluster(resource_root, name, version, fullVersion)
    return types.call(resource_root.post, CLUSTERS_PATH, ApiCluster, True,
                      data=[apicluster], api_version=api_version)[0]


def get_cluster(resource_root, name):
    """Lookup a cluster by name

    :param resource_root: The root Resource object.
    :param name: Cluster name
    :return: An ApiCluster object
    """
    return types.call(resource_root.get, "%s/%s"
                      % (CLUSTERS_PATH, name), ApiCluster)


def get_all_clusters(resource_root, view=None):
    """Get all clusters

    :param resource_root: The root Resource object.
    :return: A list of ApiCluster objects.
    """
    return types.call(resource_root.get, CLUSTERS_PATH, ApiCluster, True,
                      params=(dict(view=view) if view else None))


class ApiCluster(types.BaseApiResource):
    _ATTRIBUTES = {
        'name': None,
        'clusterUrl': None,
        'displayName': None,
        'version': None,
        'fullVersion': None,
        'hostsUrl': types.ROAttr(),
        'maintenanceMode': types.ROAttr(),
        'maintenanceOwners': types.ROAttr(),
        'entityStatus': types.ROAttr(),
    }

    def __init__(self, resource_root, name=None, version=None,
                 fullVersion=None):
        types.BaseApiObject.init(self, resource_root, locals())

    def _path(self):
        return "%s/%s" % (CLUSTERS_PATH, self.name)

    def get_service_types(self):
        """Get all service types supported by this cluster

        :return: A list of service types (strings)
        """
        resp = self._get_resource_root().get(self._path() + '/serviceTypes')
        return resp[types.ApiList.LIST_KEY]

    def get_commands(self, view=None):
        """Retrieve a list of running commands for this cluster

        :param view: View to materialize ('full' or 'summary')
        :return: A list of running commands.
        """
        return self._get("commands", types.ApiCommand, True,
                         params=(dict(view=view) if view else None))

    def create_service(self, name, service_type):
        """Create a service

        :param name: Service name
        :param service_type: Service type
        :return: An ApiService object
        """
        return services.create_service(self._get_resource_root(), name,
                                       service_type, self.name)

    def get_service(self, name):
        """Lookup a service by name

        :param name: Service name
        :return: An ApiService object
        """
        return services.get_service(self._get_resource_root(),
                                    name, self.name)

    def start(self):
        """Start all services in a cluster, respecting dependencies

        :return: Reference to the submitted command.
        """
        return self._cmd('start')

    def restart(self, restart_only_stale_services=None,
                redeploy_client_configuration=None,
                restart_service_names=None):
        """Restart all services in the cluster. Services are restarted in the

        appropriate order given their dependencies.
        :param restart_only_stale_services: Only restart services that
        have stale configuration and their dependent
        services. Default is False.
        :param redeploy_client_configuration: Re-deploy client configuration
        for all services in the cluster. Default is False.
        :param restart_service_names: Only restart services that are specified
        and their dependent services.
        :return: Reference to the submitted command.
        """
        if self._get_resource_root().version < 6:
            return self._cmd('restart')

        args = dict()
        args['restartOnlyStaleServices'] = restart_only_stale_services
        args['redeployClientConfiguration'] = redeploy_client_configuration
        if self._get_resource_root().version >= 11:
            args['restartServiceNames'] = restart_service_names

        return self._cmd('restart', data=args, api_version=6)

    def stop(self):
        """Stop all services in a cluster, respecting dependencies

        :return: Reference to the submitted command.
        """
        return self._cmd('stop')

    def deploy_client_config(self):
        """Deploys Service client configuration to the hosts on the cluster

        :return: Reference to the submitted command.
        :since: API v2
        """
        return self._cmd('deployClientConfig')

    def first_run(self):
        """Prepare and start services in a cluster

        Perform all the steps needed to prepare each service in a
        cluster and start the services in order.

        :return: Reference to the submitted command.
        :since: API v7
        """
        return self._cmd('firstRun', None, api_version=7)

    def remove_host(self, hostId):
        """Removes the association of the host with the cluster

        :return: A ApiHostRef of the host that was removed.
        :since: API v3
        """
        return self._delete("hosts/" + hostId, types.ApiHostRef, api_version=3)

    def get_service_health_status(self):
        """Lookup a service health status by name

        :return: A dict with cluster health status
        """
        health_dict = {}
        cl_services = services.get_all_services(self._get_resource_root(),
                                                cluster_name=self.name)
        for curr in cl_services:
            health_dict[curr.name] = {
                'summary': curr.get_health_summary(),
                'checks': curr.get_health_checks_status()}
        return health_dict

    def configure_for_kerberos(self, datanode_transceiver_port=None,
                               datanode_web_port=None):
        """Command to configure the cluster to use Kerberos for authentication.

        This command will configure all relevant services on a cluster for
        Kerberos usage.  This command will trigger a GenerateCredentials
        command to create Kerberos keytabs for all roles in the cluster.
        :param datanode_transceiver_port: The HDFS DataNode transceiver port
               to use. This will be applied to all DataNode role
               configuration groups. If not specified, this will default to
               1004.
        :param datanode_web_port: The HDFS DataNode web port to use. This will
               be applied to all DataNode role configuration groups. If not
               specified, this will default to 1006.
        :return: Reference to the submitted command.
        :since: API v11
        """
        args = dict()
        if datanode_transceiver_port:
            args['datanodeTransceiverPort'] = datanode_transceiver_port
        if datanode_web_port:
            args['datanodeWebPort'] = datanode_web_port
        return self._cmd('configureForKerberos', data=args, api_version=11)



# Copyright (c) 2015 Intel Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from sahara.plugins import exceptions as e
from sahara_plugin_cdh.i18n import _


class CMApiVersionError(e.SaharaPluginException):
    """Exception indicating that CM API Version does not meet requirement.

    A message indicating the reason for failure must be provided.
    """

    base_message = _("CM API version not meet requirement: %s")

    def __init__(self, message):
        self.code = "CM_API_VERSION_ERROR"
        self.message = self.base_message % message

        super(CMApiVersionError, self).__init__()


class CMApiException(e.SaharaPluginException):
    """Exception Type from CM API Errors.

    Any error result from the CM API is converted into this exception type.
    This handles errors from the HTTP level as well as the API level.
    """

    base_message = _("CM API error: %s")

    def __init__(self, message):
        self.code = "CM_API_EXCEPTION"
        self.message = self.base_message % message

        super(CMApiException, self).__init__()


class CMApiAttributeError(e.SaharaPluginException):
    """Exception indicating a CM API attribute error.

    A message indicating the reason for failure must be provided.
    """

    base_message = _("CM API attribute error: %s")

    def __init__(self, message):
        self.code = "CM_API_ATTRIBUTE_ERROR"
        self.message = self.base_message % message

        super(CMApiAttributeError, self).__init__()


class CMApiValueError(e.SaharaPluginException):
    """Exception indicating a CM API value error.

    A message indicating the reason for failure must be provided.
    """

    base_message = _("CM API value error: %s")

    def __init__(self, message):
        self.code = "CM_API_VALUE_ERROR"
        self.message = self.base_message % message

        super(CMApiValueError, self).__init__()



# Copyright (c) 2015 Intel Corporation
# Copyright (c) 2015 ISPRAS
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from sahara_plugin_cdh.plugins.cdh import abstractversionhandler as avm
from sahara_plugin_cdh.plugins.cdh.v5_7_0 import cloudera_utils
from sahara_plugin_cdh.plugins.cdh.v5_7_0 import config_helper
from sahara_plugin_cdh.plugins.cdh.v5_7_0 import deploy
from sahara_plugin_cdh.plugins.cdh.v5_7_0 import edp_engine
from sahara_plugin_cdh.plugins.cdh.v5_7_0 import images
from sahara_plugin_cdh.plugins.cdh.v5_7_0 import plugin_utils
from sahara_plugin_cdh.plugins.cdh.v5_7_0 import validation


class VersionHandler(avm.BaseVersionHandler):

    def __init__(self):
        super(VersionHandler, self).__init__()
        self.config_helper = config_helper.ConfigHelperV570()
        self.cloudera_utils = cloudera_utils.ClouderaUtilsV570()
        self.plugin_utils = plugin_utils.PluginUtilsV570()
        self.deploy = deploy
        self.edp_engine = edp_engine
        self.images = images
        self.validation = validation.ValidatorV570()



# Copyright (c) 2014 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from sahara.plugins import edp
from sahara.plugins import exceptions as pl_ex
from sahara.plugins import kerberos
from sahara.plugins import utils as u
from sahara_plugin_cdh.i18n import _


class EdpOozieEngine(edp.PluginsOozieJobEngine):

    def __init__(self, cluster):
        super(EdpOozieEngine, self).__init__(cluster)
        # will be defined in derived classes
        self.cloudera_utils = None

    def get_client(self):
        if kerberos.is_kerberos_security_enabled(self.cluster):
            return super(EdpOozieEngine, self).get_remote_client()
        return super(EdpOozieEngine, self).get_client()

    def get_hdfs_user(self):
        return 'hdfs'

    def create_hdfs_dir(self, remote, dir_name):
        edp.create_dir_hadoop2(remote, dir_name, self.get_hdfs_user())

    def get_oozie_server_uri(self, cluster):
        oozie_ip = self.cloudera_utils.pu.get_oozie(cluster).management_ip
        return 'http://%s:11000/oozie' % oozie_ip

    def get_name_node_uri(self, cluster):
        if len(self.cloudera_utils.pu.get_jns(cluster)) > 0:
            return 'hdfs://%s' % self.cloudera_utils.NAME_SERVICE
        else:
            namenode_ip = self.cloudera_utils.pu.get_namenode(cluster).fqdn()
            return 'hdfs://%s:8020' % namenode_ip

    def get_resource_manager_uri(self, cluster):
        resourcemanager = self.cloudera_utils.pu.get_resourcemanager(cluster)
        return '%s:8032' % resourcemanager.fqdn()

    def get_oozie_server(self, cluster):
        return self.cloudera_utils.pu.get_oozie(cluster)

    def validate_job_execution(self, cluster, job, data):
        oo_count = u.get_instances_count(cluster, 'OOZIE_SERVER')
        if oo_count != 1:
            raise pl_ex.InvalidComponentCountException(
                'OOZIE_SERVER', '1', oo_count)

        super(EdpOozieEngine, self).validate_job_execution(cluster, job, data)


class EdpSparkEngine(edp.PluginsSparkJobEngine):

    edp_base_version = ""

    def __init__(self, cluster):
        super(EdpSparkEngine, self).__init__(cluster)
        self.master = u.get_instance(cluster, "SPARK_YARN_HISTORY_SERVER")
        self.plugin_params["spark-user"] = "sudo -u spark "
        self.plugin_params["spark-submit"] = "spark-submit"
        self.plugin_params["deploy-mode"] = "cluster"
        self.plugin_params["master"] = "yarn-cluster"
        driver_cp = u.get_config_value_or_default(
            "Spark", "Executor extra classpath", self.cluster)
        self.plugin_params["driver-class-path"] = driver_cp

    @classmethod
    def edp_supported(cls, version):
        return version >= cls.edp_base_version

    def validate_job_execution(self, cluster, job, data):
        if not self.edp_supported(cluster.hadoop_version):
            raise pl_ex.PluginInvalidDataException(
                _('Cloudera {base} or higher required to run {type}'
                  'jobs').format(base=self.edp_base_version, type=job.type))

        shs_count = u.get_instances_count(
            cluster, 'SPARK_YARN_HISTORY_SERVER')
        if shs_count != 1:
            raise pl_ex.InvalidComponentCountException(
                'SPARK_YARN_HISTORY_SERVER', '1', shs_count)

        super(EdpSparkEngine, self).validate_job_execution(
            cluster, job, data)



# Copyright (c) 2014 Intel Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# This file only contains utils not related to cm_api, while in
# cloudera_utils the functions are cm_api involved.

import os
import telnetlib  # nosec

from oslo_log import log as logging

from sahara.plugins import context
from sahara.plugins import edp
from sahara.plugins import exceptions as exc
from sahara.plugins import recommendations_utils as ru
from sahara.plugins import resource as res
from sahara.plugins import swift_helper
from sahara.plugins import utils as u
from sahara_plugin_cdh.i18n import _
from sahara_plugin_cdh.plugins.cdh import commands as cmd
from sahara_plugin_cdh.plugins.cdh import db_helper as dh


PATH_TO_CORE_SITE_XML = '/etc/hadoop/conf/core-site.xml'
HADOOP_LIB_DIR = '/usr/lib/hadoop-mapreduce'

CM_API_PORT = 7180

LOG = logging.getLogger(__name__)

AUTO_CONFIGURATION_SCHEMA = {
    'node_configs': {
        'yarn.scheduler.minimum-allocation-mb': (
            'RESOURCEMANAGER', 'yarn_scheduler_minimum_allocation_mb'),
        'mapreduce.reduce.memory.mb': (
            'YARN_GATEWAY', 'mapreduce_reduce_memory_mb'),
        'mapreduce.map.memory.mb': (
            'YARN_GATEWAY', 'mapreduce_map_memory_mb',),
        'yarn.scheduler.maximum-allocation-mb': (
            'RESOURCEMANAGER', 'yarn_scheduler_maximum_allocation_mb'),
        'yarn.app.mapreduce.am.command-opts': (
            'YARN_GATEWAY', 'yarn_app_mapreduce_am_command_opts'),
        'yarn.nodemanager.resource.memory-mb': (
            'NODEMANAGER', 'yarn_nodemanager_resource_memory_mb'),
        'mapreduce.task.io.sort.mb': (
            'YARN_GATEWAY', 'io_sort_mb'),
        'mapreduce.map.java.opts': (
            'YARN_GATEWAY', 'mapreduce_map_java_opts'),
        'mapreduce.reduce.java.opts': (
            'YARN_GATEWAY', 'mapreduce_reduce_java_opts'),
        'yarn.app.mapreduce.am.resource.mb': (
            'YARN_GATEWAY', 'yarn_app_mapreduce_am_resource_mb')
    },
    'cluster_configs': {
        'dfs.replication': ('HDFS', 'dfs_replication')
    }
}


class CDHPluginAutoConfigsProvider(ru.HadoopAutoConfigsProvider):
    def get_datanode_name(self):
        return 'HDFS_DATANODE'


class AbstractPluginUtils(object):

    def __init__(self):
        # c_helper will be defined in derived classes.
        self.c_helper = None

    def get_role_name(self, instance, service):
        # NOTE: role name must match regexp "[_A-Za-z][-_A-Za-z0-9]{0,63}"
        shortcuts = {
            'AGENT': 'A',
            'ALERTPUBLISHER': 'AP',
            'CATALOGSERVER': 'ICS',
            'DATANODE': 'DN',
            'EVENTSERVER': 'ES',
            'HBASE_INDEXER': 'LHBI',
            'HIVEMETASTORE': 'HVM',
            'HIVESERVER2': 'HVS',
            'HOSTMONITOR': 'HM',
            'IMPALAD': 'ID',
            'JOBHISTORY': 'JS',
            'JOURNALNODE': 'JN',
            'KAFKA_BROKER': 'KB',
            'KMS': 'KMS',
            'MASTER': 'M',
            'NAMENODE': 'NN',
            'NODEMANAGER': 'NM',
            'OOZIE_SERVER': 'OS',
            'REGIONSERVER': 'RS',
            'RESOURCEMANAGER': 'RM',
            'SECONDARYNAMENODE': 'SNN',
            'SENTRY_SERVER': 'SNT',
            'SERVER': 'S',
            'SERVICEMONITOR': 'SM',
            'SOLR_SERVER': 'SLR',
            'SPARK_YARN_HISTORY_SERVER': 'SHS',
            'SQOOP_SERVER': 'S2S',
            'STATESTORE': 'ISS',
            'WEBHCAT': 'WHC',
            'HDFS_GATEWAY': 'HG',
            'YARN_GATEWAY': 'YG'
        }
        return '%s_%s' % (shortcuts.get(service, service),
                          instance.hostname().replace('-', '_'))

    def get_manager(self, cluster):
        return u.get_instance(cluster, 'CLOUDERA_MANAGER')

    def get_namenode(self, cluster):
        return u.get_instance(cluster, "HDFS_NAMENODE")

    def get_datanodes(self, cluster):
        return u.get_instances(cluster, 'HDFS_DATANODE')

    def get_hdfs_nodes(self, cluster, instances=None):
        instances = instances if instances else u.get_instances(cluster)
        return u.instances_with_services(
            instances, ["HDFS_DATANODE", "HDFS_NAMENODE",
                        "HDFS_SECONDARYNAMENODE"])

    def get_secondarynamenode(self, cluster):
        return u.get_instance(cluster, 'HDFS_SECONDARYNAMENODE')

    def get_historyserver(self, cluster):
        return u.get_instance(cluster, 'YARN_JOBHISTORY')

    def get_resourcemanager(self, cluster):
        return u.get_instance(cluster, 'YARN_RESOURCEMANAGER')

    def get_nodemanagers(self, cluster):
        return u.get_instances(cluster, 'YARN_NODEMANAGER')

    def get_oozie(self, cluster):
        return u.get_instance(cluster, 'OOZIE_SERVER')

    def get_hive_metastore(self, cluster):
        return u.get_instance(cluster, 'HIVE_METASTORE')

    def get_hive_servers(self, cluster):
        return u.get_instances(cluster, 'HIVE_SERVER2')

    def get_hue(self, cluster):
        return u.get_instance(cluster, 'HUE_SERVER')

    def get_spark_historyserver(self, cluster):
        return u.get_instance(cluster, 'SPARK_YARN_HISTORY_SERVER')

    def get_zookeepers(self, cluster):
        return u.get_instances(cluster, 'ZOOKEEPER_SERVER')

    def get_hbase_master(self, cluster):
        return u.get_instance(cluster, 'HBASE_MASTER')

    def get_sentry(self, cluster):
        return u.get_instance(cluster, 'SENTRY_SERVER')

    def get_flumes(self, cluster):
        return u.get_instances(cluster, 'FLUME_AGENT')

    def get_solrs(self, cluster):
        return u.get_instances(cluster, 'SOLR_SERVER')

    def get_sqoop(self, cluster):
        return u.get_instance(cluster, 'SQOOP_SERVER')

    def get_hbase_indexers(self, cluster):
        return u.get_instances(cluster, 'KEY_VALUE_STORE_INDEXER')

    def get_catalogserver(self, cluster):
        return u.get_instance(cluster, 'IMPALA_CATALOGSERVER')

    def get_statestore(self, cluster):
        return u.get_instance(cluster, 'IMPALA_STATESTORE')

    def get_impalads(self, cluster):
        return u.get_instances(cluster, 'IMPALAD')

    def get_kms(self, cluster):
        return u.get_instances(cluster, 'KMS')

    def get_jns(self, cluster):
        return u.get_instances(cluster, 'HDFS_JOURNALNODE')

    def get_stdb_rm(self, cluster):
        return u.get_instance(cluster, 'YARN_STANDBYRM')

    def get_kafka_brokers(self, cluster):
        return u.get_instances(cluster, 'KAFKA_BROKER')

    def convert_process_configs(self, configs):
        p_dict = {
            "CLOUDERA": ['MANAGER'],
            "NAMENODE": ['NAMENODE'],
            "DATANODE": ['DATANODE'],
            "SECONDARYNAMENODE": ['SECONDARYNAMENODE'],
            "RESOURCEMANAGER": ['RESOURCEMANAGER'],
            "NODEMANAGER": ['NODEMANAGER'],
            "JOBHISTORY": ['JOBHISTORY'],
            "OOZIE": ['OOZIE_SERVER'],
            "HIVESERVER": ['HIVESERVER2'],
            "HIVEMETASTORE": ['HIVEMETASTORE'],
            "WEBHCAT": ['WEBHCAT'],
            "HUE": ['HUE_SERVER'],
            "SPARK_ON_YARN": ['SPARK_YARN_HISTORY_SERVER'],
            "ZOOKEEPER": ['SERVER'],
            "MASTER": ['MASTER'],
            "REGIONSERVER": ['REGIONSERVER'],
            "FLUME": ['AGENT'],
            "CATALOGSERVER": ['CATALOGSERVER'],
            "STATESTORE": ['STATESTORE'],
            "IMPALAD": ['IMPALAD'],
            "KS_INDEXER": ['HBASE_INDEXER'],
            "SENTRY": ['SENTRY_SERVER'],
            "SOLR": ['SOLR_SERVER'],
            "SQOOP": ['SQOOP_SERVER'],
            "KMS": ['KMS'],
            "YARN_GATEWAY": ['YARN_GATEWAY'],
            "HDFS_GATEWAY": ['HDFS_GATEWAY'],
            "JOURNALNODE": ['JOURNALNODE'],
            "KAFKA": ['KAFKA_BROKER']
        }

        if res.is_resource_instance(configs):
            configs = configs.to_dict()
        for k in configs.keys():
            if k in p_dict.keys():
                item = configs[k]
                del configs[k]
                newkey = p_dict[k][0]
                configs[newkey] = item
        return res.create_resource(configs)

    def convert_role_showname(self, showname):
        # Yarn ResourceManager and Standby ResourceManager will
        # be converted to ResourceManager.
        name_dict = {
            'CLOUDERA_MANAGER': 'MANAGER',
            'HDFS_NAMENODE': 'NAMENODE',
            'HDFS_DATANODE': 'DATANODE',
            'HDFS_JOURNALNODE': 'JOURNALNODE',
            'HDFS_SECONDARYNAMENODE': 'SECONDARYNAMENODE',
            'YARN_RESOURCEMANAGER': 'RESOURCEMANAGER',
            'YARN_STANDBYRM': 'RESOURCEMANAGER',
            'YARN_NODEMANAGER': 'NODEMANAGER',
            'YARN_JOBHISTORY': 'JOBHISTORY',
            'OOZIE_SERVER': 'OOZIE_SERVER',
            'HIVE_SERVER2': 'HIVESERVER2',
            'HIVE_METASTORE': 'HIVEMETASTORE',
            'HIVE_WEBHCAT': 'WEBHCAT',
            'HUE_SERVER': 'HUE_SERVER',
            'SPARK_YARN_HISTORY_SERVER': 'SPARK_YARN_HISTORY_SERVER',
            'ZOOKEEPER_SERVER': 'SERVER',
            'HBASE_MASTER': 'MASTER',
            'HBASE_REGIONSERVER': 'REGIONSERVER',
            'FLUME_AGENT': 'AGENT',
            'IMPALA_CATALOGSERVER': 'CATALOGSERVER',
            'IMPALA_STATESTORE': 'STATESTORE',
            'IMPALAD': 'IMPALAD',
            'KEY_VALUE_STORE_INDEXER': 'HBASE_INDEXER',
            'SENTRY_SERVER': 'SENTRY_SERVER',
            'SOL_SERVER': 'SOLR_SERVER',
            'SQOOP_SERVER': 'SQOOP_SERVER',
        }
        return name_dict.get(showname, showname)

    def install_packages(self, instances, packages):
        # instances non-empty
        u.add_provisioning_step(
            instances[0].cluster_id, _("Install packages"), len(instances))

        with context.PluginsThreadGroup() as tg:
            for i in instances:
                tg.spawn('cdh-inst-pkgs-%s' % i.instance_name,
                         self._install_pkgs, i, packages)

    @u.event_wrapper(True)
    def _install_pkgs(self, instance, packages):
        with instance.remote() as r:
            cmd.install_packages(r, packages)

    def start_cloudera_agents(self, instances):
        # instances non-empty
        u.add_provisioning_step(
            instances[0].cluster_id, _("Start Cloudera Agents"),
            len(instances))

        with context.PluginsThreadGroup() as tg:
            for i in instances:
                tg.spawn('cdh-agent-start-%s' % i.instance_name,
                         self._start_cloudera_agent, i)

    @u.event_wrapper(True)
    def _start_cloudera_agent(self, instance):
        mng_hostname = self.get_manager(instance.cluster).hostname()
        with instance.remote() as r:
            cmd.configure_agent(r, mng_hostname)
            cmd.start_agent(r)

    def configure_swift(self, cluster, instances=None):
        if self.c_helper.is_swift_enabled(cluster):
            if not instances:
                instances = u.get_instances(cluster)
            u.add_provisioning_step(
                cluster.id, _("Configure Swift"), len(instances))

            with context.PluginsThreadGroup() as tg:
                for i in instances:
                    tg.spawn('cdh-swift-conf-%s' % i.instance_name,
                             self._configure_swift_to_inst, i)
            swift_helper.install_ssl_certs(instances)

    @u.event_wrapper(True)
    def _configure_swift_to_inst(self, instance):
        cluster = instance.cluster
        swift_lib_remote_url = self.c_helper.get_swift_lib_url(cluster)
        with instance.remote() as r:
            if r.execute_command('ls %s/hadoop-openstack.jar' % HADOOP_LIB_DIR,
                                 raise_when_error=False)[0] != 0:
                r.execute_command('sudo curl %s -o %s/hadoop-openstack.jar' % (
                    swift_lib_remote_url, HADOOP_LIB_DIR))

    def configure_sentry(self, cluster):
        manager = self.get_manager(cluster)
        with manager.remote() as r:
            dh.create_sentry_database(cluster, r)

    def put_hive_hdfs_xml(self, cluster):
        servers = self.get_hive_servers(cluster)
        with servers[0].remote() as r:
            conf_path = edp.get_hive_shared_conf_path('hdfs')
            r.execute_command(
                'sudo su - -c "hadoop fs -mkdir -p %s" hdfs'
                % os.path.dirname(conf_path))
            r.execute_command(
                'sudo su - -c "hadoop fs -put /etc/hive/conf/hive-site.xml '
                '%s" hdfs' % conf_path)

    def configure_hive(self, cluster):
        manager = self.get_manager(cluster)
        with manager.remote() as r:
            dh.create_hive_database(cluster, r)

    def install_extjs(self, cluster):
        extjs_remote_location = self.c_helper.get_extjs_lib_url(cluster)
        extjs_vm_location_dir = '/var/lib/oozie'
        extjs_vm_location_path = extjs_vm_location_dir + '/extjs.zip'
        with self.get_oozie(cluster).remote() as r:
            if r.execute_command('ls %s/ext-2.2' % extjs_vm_location_dir,
                                 raise_when_error=False)[0] != 0:
                r.execute_command('curl -L -o \'%s\' %s' % (
                    extjs_vm_location_path, extjs_remote_location),
                    run_as_root=True)
                r.execute_command('unzip %s -d %s' % (
                    extjs_vm_location_path, extjs_vm_location_dir),
                    run_as_root=True)

    def _check_cloudera_manager_started(self, manager):
        try:
            conn = telnetlib.Telnet(manager.management_ip, CM_API_PORT)
            conn.close()
            return True
        except IOError:
            return False

    @u.event_wrapper(
        True, step=_("Start Cloudera Manager"), param=('cluster', 1))
    def _start_cloudera_manager(self, cluster, timeout_config):
        manager = self.get_manager(cluster)
        with manager.remote() as r:
            cmd.start_cloudera_db(r)
            cmd.start_manager(r)
        u.plugin_option_poll(
            cluster, self._check_cloudera_manager_started, timeout_config,
            _("Await starting Cloudera Manager"), 2, {'manager': manager})

    def configure_os(self, instances):
        # instances non-empty
        u.add_provisioning_step(
            instances[0].cluster_id, _("Configure OS"), len(instances))
        with context.PluginsThreadGroup() as tg:
            for inst in instances:
                tg.spawn('cdh-repo-conf-%s' % inst.instance_name,
                         self._configure_repo_from_inst, inst)

    @u.event_wrapper(True)
    def _configure_repo_from_inst(self, instance):
        LOG.debug("Configure repos from instance {instance}".format(
                  instance=instance.instance_name))
        cluster = instance.cluster

        with instance.remote() as r:
            if cmd.is_ubuntu_os(r):
                cdh5_key = (
                    self.c_helper.get_cdh5_key_url(cluster) or
                    self.c_helper.DEFAULT_CDH5_UBUNTU_REPO_KEY_URL)
                cm5_key = (
                    self.c_helper.get_cm5_key_url(cluster) or
                    self.c_helper.DEFAULT_CM5_UBUNTU_REPO_KEY_URL)
                if self.c_helper.is_keytrustee_available():
                    kms_key = (
                        self.c_helper.get_kms_key_url(cluster) or
                        self.c_helper.DEFAULT_KEY_TRUSTEE_UBUNTU_REPO_KEY_URL)
                    kms_repo_url = self.c_helper.KEY_TRUSTEE_UBUNTU_REPO_URL
                    cmd.add_ubuntu_repository(r, kms_repo_url, 'kms')
                    cmd.add_apt_key(r, kms_key)

                cdh5_repo_content = self.c_helper.CDH5_UBUNTU_REPO
                cm5_repo_content = self.c_helper.CM5_UBUNTU_REPO

                cmd.write_ubuntu_repository(r, cdh5_repo_content, 'cdh')
                cmd.add_apt_key(r, cdh5_key)
                cmd.write_ubuntu_repository(r, cm5_repo_content, 'cm')
                cmd.add_apt_key(r, cm5_key)
                cmd.update_repository(r)

            if cmd.is_centos_os(r):
                cdh5_repo_content = self.c_helper.CDH5_CENTOS_REPO
                cm5_repo_content = self.c_helper.CM5_CENTOS_REPO
                if self.c_helper.is_keytrustee_available():
                    kms_repo_url = self.c_helper.KEY_TRUSTEE_CENTOS_REPO_URL
                    cmd.add_centos_repository(r, kms_repo_url, 'kms')

                cmd.write_centos_repository(r, cdh5_repo_content, 'cdh')
                cmd.write_centos_repository(r, cm5_repo_content, 'cm')
                cmd.update_repository(r)

    def _get_config_value(self, service, name, configs, cluster=None):
        if cluster:
            conf = cluster.cluster_configs
            if service in conf and name in conf[service]:
                return u.transform_to_num(conf[service][name])
            for node_group in cluster.node_groups:
                conf = node_group.node_configs
                if service in conf and name in conf[service]:
                    return u.transform_to_num(conf[service][name])
        for config in configs:
            if config.applicable_target == service and config.name == name:
                return u.transform_to_num(config.default_value)
        raise exc.InvalidDataException(
            _("Unable to find config: applicable_target: {target}, name: "
              "{name}").format(target=service, name=name))

    def recommend_configs(self, cluster, plugin_configs, scaling):
        provider = CDHPluginAutoConfigsProvider(
            AUTO_CONFIGURATION_SCHEMA, plugin_configs, cluster, scaling)
        provider.apply_recommended_configs()

    def start_cloudera_manager(self, cluster):
        self._start_cloudera_manager(
            cluster, self.c_helper.AWAIT_MANAGER_STARTING_TIMEOUT)

    def get_config_value(self, service, name, cluster=None):
        configs = self.c_helper.get_plugin_configs()
        return self._get_config_value(service, name, configs, cluster)



# Copyright (c) 2014 Mirantis, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc

from sahara.plugins import conductor
from sahara.plugins import context
from sahara.plugins import kerberos
from sahara_plugin_cdh.plugins.cdh import db_helper as dh
from sahara_plugin_cdh.plugins.cdh import health


class AbstractVersionHandler(object, metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def get_node_processes(self):
        return

    @abc.abstractmethod
    def get_plugin_configs(self):
        return

    @abc.abstractmethod
    def configure_cluster(self, cluster):
        return

    @abc.abstractmethod
    def start_cluster(self, cluster):
        return

    @abc.abstractmethod
    def validate(self, cluster):
        return

    @abc.abstractmethod
    def scale_cluster(self, cluster, instances):
        return

    @abc.abstractmethod
    def decommission_nodes(self, cluster, instances):
        return

    @abc.abstractmethod
    def validate_scaling(self, cluster, existing, additional):
        return

    @abc.abstractmethod
    def get_edp_engine(self, cluster, job_type):
        return

    @abc.abstractmethod
    def get_edp_job_types(self):
        return []

    @abc.abstractmethod
    def get_edp_config_hints(self, job_type):
        return {}

    @abc.abstractmethod
    def get_open_ports(self, node_group):
        return

    def on_terminate_cluster(self, cluster):
        dh.delete_passwords_from_keymanager(cluster)

    @abc.abstractmethod
    def get_image_arguments(self):
        return NotImplemented

    @abc.abstractmethod
    def pack_image(self, hadoop_version, remote, test_only=False,
                   image_arguments=None):
        pass

    @abc.abstractmethod
    def validate_images(self, cluster, test_only=False, image_arguments=None):
        pass


class BaseVersionHandler(AbstractVersionHandler):

    def __init__(self):
        # Need to be specified in subclass
        self.config_helper = None  # config helper
        self.cloudera_utils = None  # ClouderaUtils
        self.deploy = None  # to deploy
        self.edp_engine = None
        self.plugin_utils = None  # PluginUtils
        self.validation = None  # to validate

    def get_plugin_configs(self):
        result = self.config_helper.get_plugin_configs()
        result.extend(kerberos.get_config_list())
        return result

    def get_node_processes(self):
        return {
            "CLOUDERA": ['CLOUDERA_MANAGER'],
            "HDFS": ['HDFS_NAMENODE', 'HDFS_DATANODE',
                     'HDFS_SECONDARYNAMENODE', 'HDFS_JOURNALNODE'],
            "YARN": ['YARN_RESOURCEMANAGER', 'YARN_NODEMANAGER',
                     'YARN_JOBHISTORY', 'YARN_STANDBYRM'],
            "OOZIE": ['OOZIE_SERVER'],
            "HIVE": ['HIVE_SERVER2', 'HIVE_METASTORE', 'HIVE_WEBHCAT'],
            "HUE": ['HUE_SERVER'],
            "SPARK_ON_YARN": ['SPARK_YARN_HISTORY_SERVER'],
            "ZOOKEEPER": ['ZOOKEEPER_SERVER'],
            "HBASE": ['HBASE_MASTER', 'HBASE_REGIONSERVER'],
            "FLUME": ['FLUME_AGENT'],
            "IMPALA": ['IMPALA_CATALOGSERVER', 'IMPALA_STATESTORE', 'IMPALAD'],
            "KS_INDEXER": ['KEY_VALUE_STORE_INDEXER'],
            "SOLR": ['SOLR_SERVER'],
            "SQOOP": ['SQOOP_SERVER'],
            "SENTRY": ['SENTRY_SERVER'],
            "KMS": ['KMS'],
            "KAFKA": ['KAFKA_BROKER'],

            "YARN_GATEWAY": [],
            "RESOURCEMANAGER": [],
            "NODEMANAGER": [],
            "JOBHISTORY": [],

            "HDFS_GATEWAY": [],
            'DATANODE': [],
            'NAMENODE': [],
            'SECONDARYNAMENODE': [],
            'JOURNALNODE': [],

            'REGIONSERVER': [],
            'MASTER': [],

            'HIVEMETASTORE': [],
            'HIVESERVER': [],
            'WEBCAT': [],

            'CATALOGSERVER': [],
            'STATESTORE': [],
            'IMPALAD': [],
            'Kerberos': [],
        }

    def validate(self, cluster):
        self.validation.validate_cluster_creating(cluster)

    def configure_cluster(self, cluster):
        self.deploy.configure_cluster(cluster)
        conductor.cluster_update(
            context.ctx(), cluster, {
                'info':
                self.cloudera_utils.get_cloudera_manager_info(cluster)})

    def start_cluster(self, cluster):
        self.deploy.start_cluster(cluster)

        self._set_cluster_info(cluster)

    def decommission_nodes(self, cluster, instances):
        self.deploy.decommission_cluster(cluster, instances)

    def validate_scaling(self, cluster, existing, additional):
        self.validation.validate_existing_ng_scaling(cluster, existing)
        self.validation.validate_additional_ng_scaling(cluster, additional)

    def scale_cluster(self, cluster, instances):
        self.deploy.scale_cluster(cluster, instances)

    def _set_cluster_info(self, cluster):
        info = self.cloudera_utils.get_cloudera_manager_info(cluster)
        hue = self.cloudera_utils.pu.get_hue(cluster)
        if hue:
            info['Hue Dashboard'] = {
                'Web UI': 'http://%s:8888' % hue.get_ip_or_dns_name()
            }

        ctx = context.ctx()
        conductor.cluster_update(ctx, cluster, {'info': info})

    def get_edp_engine(self, cluster, job_type):
        oozie_type = self.edp_engine.EdpOozieEngine.get_supported_job_types()
        spark_type = self.edp_engine.EdpSparkEngine.get_supported_job_types()
        if job_type in oozie_type:
            return self.edp_engine.EdpOozieEngine(cluster)
        if job_type in spark_type:
            return self.edp_engine.EdpSparkEngine(cluster)
        return None

    def get_edp_job_types(self):
        return (self.edp_engine.EdpOozieEngine.get_supported_job_types() +
                self.edp_engine.EdpSparkEngine.get_supported_job_types())

    def get_edp_config_hints(self, job_type):
        return self.edp_engine.EdpOozieEngine.get_possible_job_config(job_type)

    def get_open_ports(self, node_group):
        return self.deploy.get_open_ports(node_group)

    def recommend_configs(self, cluster, scaling):
        self.plugin_utils.recommend_configs(
            cluster, self.get_plugin_configs(), scaling)

    def get_health_checks(self, cluster):
        return health.get_health_checks(cluster, self.cloudera_utils)

    def get_image_arguments(self):
        if hasattr(self, 'images'):
            return self.images.get_image_arguments()
        else:
            return NotImplemented

    def pack_image(self, hadoop_version, remote, test_only=False,
                   image_arguments=None):
        if hasattr(self, 'images'):
            self.images.pack_image(
                remote, test_only=test_only, image_arguments=image_arguments)

    def validate_images(self, cluster, test_only=False, image_arguments=None):
        if hasattr(self, 'images'):
            self.images.validate_images(
                cluster, test_only=test_only, image_arguments=image_arguments)



# Copyright (c) 2015 Intel Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# The contents of this file are mainly copied from cm_api sources,
# released by Cloudera. Codes not used by Sahara CDH plugin are removed.
# You can find the original codes at
#
#     https://github.com/cloudera/cm_api/tree/master/python/src/cm_api
#
# To satisfy the pep8 and python3 tests, we did some changes to the codes.
# We also change some importings to use Sahara inherited classes.

from sahara_plugin_cdh.plugins.cdh.client import types

USERS_PATH = "/users"


def get_user(resource_root, username):
    """Look up a user by username.

    @param resource_root: The root Resource object
    @param username: Username to look up
    @return: An ApiUser object
    """
    return types.call(resource_root.get,
                      '%s/%s' % (USERS_PATH, username), ApiUser)


def update_user(resource_root, user):
    """Update a user.

    Replaces the user's details with those provided.

    @param resource_root: The root Resource object
    @param user: An ApiUser object
    @return: An ApiUser object
    """
    return types.call(resource_root.put,
                      '%s/%s' % (USERS_PATH, user.name), ApiUser, data=user)


class ApiUser(types.BaseApiResource):
    _ATTRIBUTES = {
        'name': None,
        'password': None,
        'roles': None,
    }

    def __init__(self, resource_root, name=None, password=None, roles=None):
        types.BaseApiObject.init(self, resource_root, locals())



# Copyright (c) 2016 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

from sahara_plugin_cdh.plugins.cdh.v5_9_0 import deploy
from sahara_plugin_cdh.tests.unit import base


class DeployCDHV590(base.SaharaTestCase):

    def setUp(self):
        super(DeployCDHV590, self).setUp()
        self.master = mock.MagicMock()
        self.master.node_group.node_processes = [
            "HDFS_NAMENODE", "YARN_RESOURCEMANAGER", "CLOUDERA_MANAGER",
            "SENTRY_SERVER", "YARN_NODEMANAGER", "ZOOKEEPER_SERVER",
            "OOZIE_SERVER", "YARN_JOBHISTORY", "HDFS_SECONDARYNAMENODE",
            "HIVE_METASTORE", "HIVE_SERVER2", "SPARK_YARN_HISTORY_SERVER",
            "HBASE_MASTER", "HBASE_REGIONSERVER", "HUE_SERVER", "KMS",
            "FLUME_AGENT", "SOLR_SERVER", "SQOOP_SERVER", "IMPALA_STATESTORE",
            "IMPALA_CATALOGSERVER", "IMPALAD", "KEY_VALUE_STORE_INDEXER",
        ]
        self.worker = mock.MagicMock()
        self.worker.node_group.node_processes = [
            "HDFS_DATANODE", "HDFS_JOURNALNODE", "JOURNALNODE",
            "YARN_NODEMANAGER", "YARN_STANDBYRM",
        ]
        self.instances = [self.master, self.worker]
        self.cluster = mock.MagicMock()

        self.is_cdh_exists = mock.patch(
            "sahara_plugin_cdh.plugins.cdh.commands.is_pre_installed_cdh",
            return_value=False)
        self.is_cdh_exists.start()
        self._create_facade = mock.patch(
            "sahara.db.sqlalchemy.api._create_facade_lazily")
        self._create_facade.start()

    def tearDown(self):
        self.is_cdh_exists.stop()
        self._create_facade.stop()
        super(DeployCDHV590, self).tearDown()

    @mock.patch("sahara.plugins.utils.get_instances")
    @mock.patch("sahara_plugin_cdh.plugins.cdh.v5_9_0.deploy.CU")
    def test_configure_cluster(self, mock_cu, mock_get_instances):
        mock_get_instances.return_value = self.instances

        deploy.configure_cluster(self.cluster)

        mock_cu.pu.configure_os.assert_called_once_with(self.instances)
        mock_cu.pu.install_packages.assert_called_once_with(self.instances,
                                                            deploy.PACKAGES)
        mock_cu.pu.start_cloudera_agents.assert_called_once_with(
            self.instances)
        mock_cu.pu.start_cloudera_manager.assert_called_once_with(self.cluster)
        mock_cu.update_cloudera_password.assert_called_once_with(self.cluster)
        mock_cu.await_agents.assert_called_once_with(self.cluster,
                                                     self.instances)
        mock_cu.create_mgmt_service.assert_called_once_with(self.cluster)
        mock_cu.create_services.assert_called_once_with(self.cluster)
        mock_cu.configure_services.assert_called_once_with(self.cluster)
        mock_cu.configure_instances.assert_called_once_with(self.instances,
                                                            self.cluster)
        mock_cu.deploy_configs.assert_called_once_with(self.cluster)

    @mock.patch("sahara_plugin_cdh.plugins.cdh.v5_9_0.deploy.CU")
    def test__start_roles(self, mock_cu):

        deploy._start_roles(self.cluster, self.instances)

        mock_cu.get_service_by_role.assert_any_call('DATANODE',
                                                    instance=self.worker)
        mock_cu.get_service_by_role.assert_any_call('NODEMANAGER',
                                                    instance=self.master)
        mock_cu.get_service_by_role.assert_any_call('NODEMANAGER',
                                                    instance=self.worker)
        self.assertEqual(mock_cu.start_roles.call_count, 3)

    @mock.patch("sahara_plugin_cdh.plugins.cdh.v5_9_0.deploy._start_roles")
    @mock.patch("sahara_plugin_cdh.plugins.cdh.v5_9_0.deploy.CU")
    def test_scale_cluster(self, mock_cu, mock__start_roles):
        deploy.scale_cluster(self.cluster, None)
        self.assertEqual(mock_cu.call_count, 0)
        self.assertEqual(mock__start_roles.call_count, 0)

        deploy.scale_cluster(self.cluster, self.instances)

        mock_cu.pu.configure_os.assert_called_once_with(self.instances)
        mock_cu.pu.install_packages.assert_called_once_with(self.instances,
                                                            deploy.PACKAGES)
        mock_cu.pu.start_cloudera_agents.assert_called_once_with(
            self.instances)
        mock_cu.await_agents.assert_called_once_with(self.cluster,
                                                     self.instances)
        mock_cu.configure_instances.assert_called_once_with(self.instances,
                                                            self.cluster)
        mock_cu.update_configs.assert_called_once_with(self.instances)
        mock_cu.pu.configure_swift.assert_called_once_with(self.cluster,
                                                           self.instances)
        mock_cu.refresh_datanodes.assert_called_once_with(self.cluster)
        mock__start_roles.assert_called_once_with(self.cluster,
                                                  self.instances)

    @mock.patch("sahara_plugin_cdh.plugins.cdh.v5_9_0.deploy.CU")
    def test_decommission_cluster(self, mock_cu):
        deploy.decommission_cluster(self.cluster, self.instances)
        dns = []
        dns_2 = []
        nms = []
        nms_2 = []
        for i in self.instances:
            if 'HDFS_DATANODE' in i.node_group.node_processes:
                dns.append(mock_cu.pu.get_role_name(i, 'DATANODE'))
                dns_2.append(mock_cu.pu.get_role_name(i, 'HDFS_GATEWAY'))
            if 'YARN_NODEMANAGER' in i.node_group.node_processes:
                nms.append(mock_cu.pu.get_role_name(i, 'NODEMANAGER'))
                nms_2.append(mock_cu.pu.get_role_name(i, 'YARN_GATEWAY'))
        mock_cu.decommission_nodes.assert_any_call(
            self.cluster, 'DATANODE', dns, dns_2)
        mock_cu.decommission_nodes.assert_any_call(
            self.cluster, 'NODEMANAGER', nms, nms_2)
        mock_cu.delete_instances.assert_called_once_with(self.cluster,
                                                         self.instances)
        mock_cu.refresh_datanodes.assert_called_once_with(self.cluster)
        mock_cu.refresh_yarn_nodes.assert_called_once_with(self.cluster)

    @mock.patch("sahara_plugin_cdh.plugins.cdh.v5_9_0.deploy.CU")
    def test__prepare_cluster(self, mock_cu):
        deploy._prepare_cluster(self.cluster)

        mock_cu.pu.install_extjs.assert_called_once_with(self.cluster)
        mock_cu.pu.configure_hive.assert_called_once_with(self.cluster)
        mock_cu.pu.configure_sentry.assert_called_once_with(self.cluster)

    @mock.patch("sahara.service.edp.hdfs_helper.create_hbase_common_lib")
    @mock.patch("sahara_plugin_cdh.plugins.cdh.v5_9_0.deploy.CU")
    def test__finish_cluster_starting(self, mock_cu, mock_create_hbase):
        deploy._finish_cluster_starting(self.cluster)
        mock_cu.pu.put_hive_hdfs_xml.assert_called_once_with(self.cluster)
        self.assertTrue(mock_create_hbase.called)
        mock_cu.start_service.assert_called_once_with(
            mock_cu.get_service_by_role('AGENT', self.cluster))

    @mock.patch("sahara_plugin_cdh.plugins.cdh.v5_9_0.deploy."
                "_finish_cluster_starting")
    @mock.patch("sahara_plugin_cdh.plugins.cdh.v5_9_0.deploy._prepare_cluster")
    @mock.patch("sahara_plugin_cdh.plugins.cdh.v5_9_0.deploy.CU")
    def test_start_cluster(self, mock_cu, mock_prepare, mock_finish):
        jns_count = 0
        for i in self.instances:
            if "HDFS_JOURNALNODE" in i.node_group.node_processes:
                jns_count += 1
        mock_cu.pu.get_jns.return_value.__len__.return_value = jns_count

        deploy.start_cluster(self.cluster)

        mock_prepare.assert_called_once_with(self.cluster)
        mock_cu.first_run.assert_called_once_with(self.cluster)
        mock_cu.pu.configure_swift.assert_called_once_with(self.cluster)
        if jns_count > 0:
            mock_cu.enable_namenode_ha.assert_called_once_with(self.cluster)
            mock_cu.update_role_config.assert_any_call(
                mock_cu.pu.get_secondarynamenode(self.cluster),
                'HDFS_NAMENODE'
            )
        mock_cu.enable_resourcemanager_ha.assert_called_once_with(self.cluster)
        mock_cu.update_role_config.assert_any_call(
            mock_cu.pu.get_stdb_rm(self.cluster), 'YARN_STANDBYRM')
        mock_finish.assert_called_once_with(self.cluster)

    def test_get_open_ports(self):
        master_ports = [
            9000,
            7180, 7182, 7183, 7432, 7184, 8084, 8086, 10101,
            9997, 9996, 8087, 9998, 9999, 8085, 9995, 9994,
            8020, 8022, 50070, 50470,
            50090, 50495,
            8030, 8031, 8032, 8033, 8088,
            8040, 8041, 8042,
            10020, 19888,
            9083,
            10000,
            8888,
            11000, 11001,
            18088,
            2181, 3181, 4181, 9010,
            60000,
            60020,
            41414,
            8038,
            8983, 8984,
            8005, 12000,
            25020, 26000,
            25010, 24000,
            21050, 21000, 23000, 25000, 28000, 22000,
            16000, 16001
        ]
        deploy.get_open_ports(self.master.node_group)
        self.assertItemsEqual(master_ports,
                              deploy.get_open_ports(self.master.node_group))
        worker_ports = [
            9000,
            50010, 1004, 50075, 1006, 50020,
            8480, 8481, 8485,
            8040, 8041, 8042,
            8030, 8031, 8032, 8033, 8088
        ]
        self.assertItemsEqual(worker_ports,
                              deploy.get_open_ports(self.worker.node_group))



# Copyright (c) 2014 Intel Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# The contents of this file are mainly copied from cm_api sources,
# released by Cloudera. Codes not used by Sahara CDH plugin are removed.
# You can find the original codes at
#
#     https://github.com/cloudera/cm_api/tree/master/python/src/cm_api
#
# To satisfy the pep8 and python3 tests, we did some changes to the codes.
# We also change some importings to use Sahara inherited classes.

from oslo_serialization import jsonutils as json

from sahara_plugin_cdh.plugins.cdh.client import role_config_groups
from sahara_plugin_cdh.plugins.cdh.client import roles
from sahara_plugin_cdh.plugins.cdh.client import types

SERVICES_PATH = "/clusters/%s/services"
SERVICE_PATH = "/clusters/%s/services/%s"
ROLETYPES_CFG_KEY = 'roleTypeConfigs'


def create_service(resource_root, name, service_type,
                   cluster_name="default"):
    """Create a service

    :param resource_root: The root Resource object.
    :param name: Service name
    :param service_type: Service type
    :param cluster_name: Cluster name
    :return: An ApiService object
    """
    apiservice = ApiService(resource_root, name, service_type)
    return types.call(resource_root.post, SERVICES_PATH % (cluster_name,),
                      ApiService, True, data=[apiservice])[0]


def get_service(resource_root, name, cluster_name="default"):
    """Lookup a service by name

    :param resource_root: The root Resource object.
    :param name: Service name
    :param cluster_name: Cluster name
    :return: An ApiService object
    """
    return _get_service(resource_root, "%s/%s"
                        % (SERVICES_PATH % (cluster_name,), name))


def _get_service(resource_root, path):
    return types.call(resource_root.get, path, ApiService)


def get_all_services(resource_root, cluster_name="default", view=None):
    """Get all services

    :param resource_root: The root Resource object.
    :param cluster_name: Cluster name
    :return: A list of ApiService objects.
    """
    return types.call(resource_root.get, SERVICES_PATH % (cluster_name,),
                      ApiService, True,
                      params=(dict(view=view) if view else None))


def delete_service(resource_root, name, cluster_name="default"):
    """Delete a service by name

    :param resource_root: The root Resource object.
    :param name: Service name
    :param cluster_name: Cluster name
    :return: The deleted ApiService object
    """
    return types.call(resource_root.delete,
                      "%s/%s" % (SERVICES_PATH % (cluster_name,), name),
                      ApiService)


class ApiService(types.BaseApiResource):
    _ATTRIBUTES = {
        'name': None,
        'type': None,
        'displayName': None,
        'serviceState': types.ROAttr(),
        'healthSummary': types.ROAttr(),
        'healthChecks': types.ROAttr(),
        'clusterRef': types.ROAttr(types.ApiClusterRef),
        'configStale': types.ROAttr(),
        'configStalenessStatus': types.ROAttr(),
        'clientConfigStalenessStatus': types.ROAttr(),
        'serviceUrl': types.ROAttr(),
        'maintenanceMode': types.ROAttr(),
        'maintenanceOwners': types.ROAttr(),
    }

    def __init__(self, resource_root, name=None, type=None):
        types.BaseApiObject.init(self, resource_root, locals())

    def __str__(self):
        return ("<ApiService>: %s (cluster: %s)"
                % (self.name, self._get_cluster_name()))

    def _get_cluster_name(self):
        if hasattr(self, 'clusterRef') and self.clusterRef:
            return self.clusterRef.clusterName
        return None

    def _path(self):
        """Return the API path for this service

        This method assumes that lack of a cluster reference means that the
        object refers to the Cloudera Management Services instance.
        """
        if self._get_cluster_name():
            return SERVICE_PATH % (self._get_cluster_name(), self.name)
        else:
            return '/cm/service'

    def _role_cmd(self, cmd, roles, api_version=1):
        return self._post("roleCommands/" + cmd, types.ApiBulkCommandList,
                          data=roles, api_version=api_version)

    def _parse_svc_config(self, json_dic, view=None):
        """Parse a json-decoded ApiServiceConfig dictionary into a 2-tuple

        :param json_dic: The json dictionary with the config data.
        :param view: View to materialize.
        :return: 2-tuple (service config dictionary, role type configurations)
        """
        svc_config = types.json_to_config(json_dic, view == 'full')
        rt_configs = {}
        if ROLETYPES_CFG_KEY in json_dic:
            for rt_config in json_dic[ROLETYPES_CFG_KEY]:
                rt_configs[rt_config['roleType']] = types.json_to_config(
                    rt_config, view == 'full')

        return (svc_config, rt_configs)

    def create_yarn_job_history_dir(self):
        """Create the Yarn job history directory

        :return: Reference to submitted command.
        :since: API v6
        """
        return self._cmd('yarnCreateJobHistoryDirCommand', api_version=6)

    def get_config(self, view=None):
        """Retrieve the service's configuration

        Retrieves both the service configuration and role type configuration
        for each of the service's supported role types. The role type
        configurations are returned as a dictionary, whose keys are the
        role type name, and values are the respective configuration
        dictionaries.

        The 'summary' view contains strings as the dictionary values. The full
        view contains types.ApiConfig instances as the values.

        :param view: View to materialize ('full' or 'summary')
        :return: 2-tuple (service config dictionary, role type configurations)
        """
        path = self._path() + '/config'
        resp = self._get_resource_root().get(
            path, params=(dict(view=view) if view else None))
        return self._parse_svc_config(resp, view)

    def update_config(self, svc_config, **rt_configs):
        """Update the service's configuration

        :param svc_config: Dictionary with service configuration to update.
        :param rt_configs: Dict of role type configurations to update.
        :return: 2-tuple (service config dictionary, role type configurations)
        """
        path = self._path() + '/config'

        if svc_config:
            data = types.config_to_api_list(svc_config)
        else:
            data = {}
        if rt_configs:
            rt_list = []
            for rt, cfg in rt_configs.items():
                rt_data = types.config_to_api_list(cfg)
                rt_data['roleType'] = rt
                rt_list.append(rt_data)
            data[ROLETYPES_CFG_KEY] = rt_list

        resp = self._get_resource_root().put(path, data=json.dumps(data))
        return self._parse_svc_config(resp)

    def create_role(self, role_name, role_type, host_id):
        """Create a role

        :param role_name: Role name
        :param role_type: Role type
        :param host_id: ID of the host to assign the role to
        :return: An ApiRole object
        """
        return roles.create_role(self._get_resource_root(), self.name,
                                 role_type, role_name, host_id,
                                 self._get_cluster_name())

    def delete_role(self, name):
        """Delete a role by name

        :param name: Role name
        :return: The deleted ApiRole object
        """
        return roles.delete_role(self._get_resource_root(), self.name, name,
                                 self._get_cluster_name())

    def get_roles_by_type(self, role_type, view=None):
        """Get all roles of a certain type in a service

        :param role_type: Role type
        :param view: View to materialize ('full' or 'summary')
        :return: A list of ApiRole objects.
        """
        return roles.get_roles_by_type(self._get_resource_root(), self.name,
                                       role_type, self._get_cluster_name(),
                                       view)

    def get_all_role_config_groups(self):
        """Get a list of role configuration groups in the service

        :return: A list of ApiRoleConfigGroup objects.
        :since: API v3
        """
        return role_config_groups.get_all_role_config_groups(
            self._get_resource_root(), self.name, self._get_cluster_name())

    def start(self):
        """Start a service

        :return: Reference to the submitted command.
        """
        return self._cmd('start')

    def stop(self):
        """Stop a service

        :return: Reference to the submitted command.
        """
        return self._cmd('stop')

    def restart(self):
        """Restart a service

        :return: Reference to the submitted command.
        """
        return self._cmd('restart')

    def get_health_summary(self):
        return getattr(self, 'healthSummary', None)

    def get_health_checks_status(self):
        return getattr(self, 'healthChecks', None)

    def start_roles(self, *role_names):
        """Start a list of roles

        :param role_names: names of the roles to start.
        :return: List of submitted commands.
        """
        return self._role_cmd('start', role_names)

    def create_hbase_root(self):
        """Create the root directory of an HBase service

        :return: Reference to the submitted command.
        """
        return self._cmd('hbaseCreateRoot')

    def create_hdfs_tmp(self):
        """Create /tmp directory in HDFS

        Create the /tmp directory in HDFS with appropriate ownership and
        permissions.

        :return: Reference to the submitted command
        :since: API v2
        """
        return self._cmd('hdfsCreateTmpDir')

    def refresh(self, *role_names):
        """Execute the "refresh" command on a set of roles

        :param role_names: Names of the roles to refresh.
        :return: Reference to the submitted command.
        """
        return self._role_cmd('refresh', role_names)

    def decommission(self, *role_names):
        """Decommission roles in a service

        :param role_names: Names of the roles to decommission.
        :return: Reference to the submitted command.
        """
        return self._cmd('decommission', data=role_names)

    def deploy_client_config(self, *role_names):
        """Deploys client configuration to the hosts where roles are running

        :param role_names: Names of the roles to decommission.
        :return: Reference to the submitted command.
        """
        return self._cmd('deployClientConfig', data=role_names)

    def format_hdfs(self, *namenodes):
        """Format NameNode instances of an HDFS service

        :param namenodes: Name of NameNode instances to format.
        :return: List of submitted commands.
        """
        return self._role_cmd('hdfsFormat', namenodes)

    def install_oozie_sharelib(self):
        """Installs the Oozie ShareLib

        Oozie must be stopped before running this command.

        :return: Reference to the submitted command.
        :since: API v3
        """
        return self._cmd('installOozieShareLib', api_version=3)

    def create_oozie_db(self):
        """Creates the Oozie Database Schema in the configured database

        :return: Reference to the submitted command.
        :since: API v2
        """
        return self._cmd('createOozieDb', api_version=2)

    def upgrade_oozie_db(self):
        """Upgrade Oozie Database schema as part of a major version upgrade

        :return: Reference to the submitted command.
        :since: API v6
        """
        return self._cmd('oozieUpgradeDb', api_version=6)

    def create_hive_metastore_tables(self):
        """Creates the Hive metastore tables in the configured database

        Will do nothing if tables already exist. Will not perform an upgrade.

        :return: Reference to the submitted command.
        :since: API v3
        """
        return self._cmd('hiveCreateMetastoreDatabaseTables', api_version=3)

    def create_hive_warehouse(self):
        """Creates the Hive warehouse directory in HDFS

        :return: Reference to the submitted command.
        :since: API v3
        """
        return self._cmd('hiveCreateHiveWarehouse')

    def create_hive_userdir(self):
        """Creates the Hive user directory in HDFS

        :return: Reference to the submitted command.
        :since: API v4
        """
        return self._cmd('hiveCreateHiveUserDir')

    def enable_nn_ha(self, active_name, standby_host_id, nameservice, jns,
                     standby_name_dir_list=None, qj_name=None,
                     standby_name=None, active_fc_name=None,
                     standby_fc_name=None, zk_service_name=None,
                     force_init_znode=True,
                     clear_existing_standby_name_dirs=True,
                     clear_existing_jn_edits_dir=True):
        """Enable High Availability (HA) with Auto-Failover for HDFS NameNode

        @param active_name: Name of Active NameNode.
        @param standby_host_id: ID of host where Standby NameNode will be
                                created.
        @param nameservice: Nameservice to be used while enabling HA.
                            Optional if Active NameNode already has this
                            config set.
        @param jns: List of Journal Nodes to be created during the command.
                    Each element of the list must be a dict containing the
                    following items:
                    - jns['jnHostId']: ID of the host where the new JournalNode
                                       will be created.
                    - jns['jnName']: Name of the JournalNode role (optional)
                    - jns['jnEditsDir']: Edits dir of the JournalNode. Can be
                                         omitted if the config is already set
                                         at RCG level.
        @param standby_name_dir_list: List of directories for the new Standby
                                      NameNode. If not provided then it will
                                      use same dirs as Active NameNode.
        @param qj_name: Name of the journal located on each JournalNodes'
                        filesystem. This can be optionally provided if the
                        config hasn't been already set for the Active NameNode.
                        If this isn't provided and Active NameNode doesn't
                        also have the config, then nameservice is used by
                        default.
        @param standby_name: Name of the Standby NameNode role to be created
                             (Optional).
        @param active_fc_name: Name of the Active Failover Controller role to
                               be created (Optional).
        @param standby_fc_name: Name of the Standby Failover Controller role to
                                be created (Optional).
        @param zk_service_name: Name of the ZooKeeper service to use for auto-
                                failover. If HDFS service already depends on a
                                ZooKeeper service then that ZooKeeper service
                                will be used for auto-failover and in that case
                                this parameter can either be omitted or should
                                be the same ZooKeeper service.
        @param force_init_znode: Indicates if the ZNode should be force
                                 initialized if it is already present. Useful
                                 while re-enabling High Availability. (Default:
                                 TRUE)
        @param clear_existing_standby_name_dirs: Indicates if the existing name
                                                 directories for Standby
                                                 NameNode should be cleared
                                                 during the workflow.
                                                 Useful while re-enabling High
                                                 Availability. (Default: TRUE)
        @param clear_existing_jn_edits_dir: Indicates if the existing edits
                                            directories for the JournalNodes
                                            for the specified nameservice
                                            should be cleared during the
                                            workflow. Useful while re-enabling
                                            High Availability. (Default: TRUE)
        @return: Reference to the submitted command.
        @since: API v6
        """
        args = dict(
            activeNnName=active_name,
            standbyNnName=standby_name,
            standbyNnHostId=standby_host_id,
            standbyNameDirList=standby_name_dir_list,
            nameservice=nameservice,
            qjName=qj_name,
            activeFcName=active_fc_name,
            standbyFcName=standby_fc_name,
            zkServiceName=zk_service_name,
            forceInitZNode=force_init_znode,
            clearExistingStandbyNameDirs=clear_existing_standby_name_dirs,
            clearExistingJnEditsDir=clear_existing_jn_edits_dir,
            jns=jns
        )
        return self._cmd('hdfsEnableNnHa', data=args, api_version=6)

    def enable_rm_ha(self, new_rm_host_id, zk_service_name=None):
        """Enable high availability for a YARN ResourceManager.

        @param new_rm_host_id: id of the host where the second ResourceManager
                               will be added.
        @param zk_service_name: Name of the ZooKeeper service to use for auto-
                                failover. If YARN service depends on a
                                ZooKeeper service then that ZooKeeper service
                                will be used for auto-failover and in that case
                                this parameter can be omitted.
        @return: Reference to the submitted command.
        @since: API v6
        """
        args = dict(
            newRmHostId=new_rm_host_id,
            zkServiceName=zk_service_name
        )
        return self._cmd('enableRmHa', data=args)


class ApiServiceSetupInfo(ApiService):
    _ATTRIBUTES = {
        'name': None,
        'type': None,
        'config': types.Attr(types.ApiConfig),
        'roles': types.Attr(roles.ApiRole),
    }

    def __init__(self, name=None, type=None,
                 config=None, roles=None):
        # The BaseApiObject expects a resource_root, which we don't care about
        resource_root = None
        # Unfortunately, the json key is called "type". So our input arg
        # needs to be called "type" as well, despite it being a python keyword.
        types.BaseApiObject.init(self, None, locals())

    def set_config(self, config):
        """Set the service configuration

        :param config: A dictionary of config key/value
        """
        if self.config is None:
            self.config = {}
        self.config.update(types.config_to_api_list(config))

    def add_role_info(self, role_name, role_type, host_id, config=None):
        """Add a role info

        The role will be created along with the service setup.

        :param role_name: Role name
        :param role_type: Role type
        :param host_id: The host where the role should run
        :param config: (Optional) A dictionary of role config values
        """
        if self.roles is None:
            self.roles = []
        api_config_list = (config is not None
                           and types.config_to_api_list(config)
                           or None)
        self.roles.append({
            'name': role_name,
            'type': role_type,
            'hostRef': {'hostId': host_id},
            'config': api_config_list})



# Copyright (c) 2016 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from sahara.plugins import provisioning as p
from sahara.plugins import utils
from sahara_plugin_cdh.plugins.cdh import config_helper as c_h


class ConfigHelperV570(c_h.ConfigHelper):
    path_to_config = 'plugins/cdh/v5_7_0/resources/'

    CDH5_UBUNTU_REPO = (
        'deb [arch=amd64] http://archive.cloudera.com/cdh5'
        '/ubuntu/trusty/amd64/cdh trusty-cdh5.7.0 contrib'
        '\ndeb-src http://archive.cloudera.com/cdh5/ubuntu'
        '/trusty/amd64/cdh trusty-cdh5.7.0 contrib')

    DEFAULT_CDH5_UBUNTU_REPO_KEY_URL = (
        'http://archive.cloudera.com/cdh5/ubuntu'
        '/trusty/amd64/cdh/archive.key')

    CM5_UBUNTU_REPO = (
        'deb [arch=amd64] http://archive.cloudera.com/cm5'
        '/ubuntu/trusty/amd64/cm trusty-cm5.7.0 contrib'
        '\ndeb-src http://archive.cloudera.com/cm5/ubuntu'
        '/trusty/amd64/cm trusty-cm5.7.0 contrib')

    DEFAULT_CM5_UBUNTU_REPO_KEY_URL = (
        'http://archive.cloudera.com/cm5/ubuntu'
        '/trusty/amd64/cm/archive.key')

    CDH5_CENTOS_REPO = (
        '[cloudera-cdh5]'
        '\nname=Cloudera\'s Distribution for Hadoop, Version 5'
        '\nbaseurl=http://archive.cloudera.com/cdh5/redhat/6'
        '/x86_64/cdh/5.7.0/'
        '\ngpgkey = http://archive.cloudera.com/cdh5/redhat/6'
        '/x86_64/cdh/RPM-GPG-KEY-cloudera'
        '\ngpgcheck = 1')

    CM5_CENTOS_REPO = (
        '[cloudera-manager]'
        '\nname=Cloudera Manager'
        '\nbaseurl=http://archive.cloudera.com/cm5/redhat/6'
        '/x86_64/cm/5.7.0/'
        '\ngpgkey = http://archive.cloudera.com/cm5/redhat/6'
        '/x86_64/cm/RPM-GPG-KEY-cloudera'
        '\ngpgcheck = 1')

    KEY_TRUSTEE_UBUNTU_REPO_URL = (
        'http://archive.cloudera.com/navigator-'
        'keytrustee5/ubuntu/trusty/amd64/navigator-'
        'keytrustee/cloudera.list')

    DEFAULT_KEY_TRUSTEE_UBUNTU_REPO_KEY_URL = (
        'http://archive.cloudera.com/navigator-'
        'keytrustee5/ubuntu/trusty/amd64/navigator-'
        'keytrustee/archive.key')

    KEY_TRUSTEE_CENTOS_REPO_URL = (
        'http://archive.cloudera.com/navigator-'
        'keytrustee5/redhat/6/x86_64/navigator-'
        'keytrustee/navigator-keytrustee5.repo')

    DEFAULT_SWIFT_LIB_URL = (
        'https://repository.cloudera.com/artifactory/repo/org'
        '/apache/hadoop/hadoop-openstack/2.6.0-cdh5.7.0'
        '/hadoop-openstack-2.6.0-cdh5.7.0.jar')

    SWIFT_LIB_URL = p.Config(
        'Hadoop OpenStack library URL', 'general', 'cluster', priority=1,
        default_value=DEFAULT_SWIFT_LIB_URL,
        description=("Library that adds Swift support to CDH. The file"
                     " will be downloaded by VMs."))

    HIVE_SERVER2_SENTRY_SAFETY_VALVE = utils.get_file_text(
        path_to_config + 'hive-server2-sentry-safety.xml',
        'sahara_plugin_cdh')

    HIVE_METASTORE_SENTRY_SAFETY_VALVE = utils.get_file_text(
        path_to_config + 'hive-metastore-sentry-safety.xml',
        'sahara_plugin_cdh')

    SENTRY_IMPALA_CLIENT_SAFETY_VALVE = utils.get_file_text(
        path_to_config + 'sentry-impala-client-safety.xml',
        'sahara_plugin_cdh')

    def __init__(self):
        super(ConfigHelperV570, self).__init__()
        self.priority_one_confs = self._load_json(
            self.path_to_config + 'priority-one-confs.json')
        self._init_all_ng_plugin_configs()



# Copyright (c) 2014 Intel Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# The contents of this file are mainly copied from cm_api sources,
# released by Cloudera. Codes not used by Sahara CDH plugin are removed.
# You can find the original codes at
#
#     https://github.com/cloudera/cm_api/tree/master/python/src/cm_api
#
# To satisfy the pep8 and python3 tests, we did some changes to the codes.
# We also change some importings to use Sahara inherited classes.

import datetime

from sahara_plugin_cdh.plugins.cdh.client import types

HOSTS_PATH = "/hosts"


def get_all_hosts(resource_root, view=None):
    """Get all hosts

    :param resource_root: The root Resource object.
    :return: A list of ApiHost objects.
    """
    return types.call(resource_root.get, HOSTS_PATH, ApiHost, True,
                      params=(dict(view=view) if view else None))


def delete_host(resource_root, host_id):
    """Delete a host by id

    :param resource_root: The root Resource object.
    :param host_id: Host id
    :return: The deleted ApiHost object
    """
    return types.call(resource_root.delete, "%s/%s"
                      % (HOSTS_PATH, host_id), ApiHost)


class ApiHost(types.BaseApiResource):
    _ATTRIBUTES = {
        'hostId': None,
        'hostname': None,
        'ipAddress': None,
        'rackId': None,
        'status': types.ROAttr(),
        'lastHeartbeat': types.ROAttr(datetime.datetime),
        'roleRefs': types.ROAttr(types.ApiRoleRef),
        'healthSummary': types.ROAttr(),
        'healthChecks': types.ROAttr(),
        'hostUrl': types.ROAttr(),
        'commissionState': types.ROAttr(),
        'maintenanceMode': types.ROAttr(),
        'maintenanceOwners': types.ROAttr(),
        'numCores': types.ROAttr(),
        'totalPhysMemBytes': types.ROAttr(),
    }

    def __init__(self, resource_root, hostId=None, hostname=None,
                 ipAddress=None, rackId=None):
        types.BaseApiObject.init(self, resource_root, locals())

    def __str__(self):
        return "<ApiHost>: %s (%s)" % (self.hostId, self.ipAddress)

    def _path(self):
        return HOSTS_PATH + '/' + self.hostId

    def put_host(self):
        """Update this resource

        note (mionkin):Currently, according to Cloudera docs,
                       only updating the rackId is supported.
                       All other fields of the host will be ignored.
        :return: The updated object.
        """
        return self._put('', ApiHost, data=self)



# Copyright (c) 2014 Intel Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# The contents of this file are mainly copied from cm_api sources,
# released by Cloudera. Codes not used by Sahara CDH plugin are removed.
# You can find the original codes at
#
#     https://github.com/cloudera/cm_api/tree/master/python/src/cm_api
#
# To satisfy the pep8 and python3 tests, we did some changes to the codes.
# We also change some importings to use Sahara inherited classes.

from sahara_plugin_cdh.plugins.cdh.client.services import ApiService
from sahara_plugin_cdh.plugins.cdh.client import types


class ClouderaManager(types.BaseApiResource):
    """The Cloudera Manager instance

    Provides access to CM configuration and services.
    """
    def __init__(self, resource_root):
        types.BaseApiObject.init(self, resource_root)

    def _path(self):
        return '/cm'

    def create_mgmt_service(self, service_setup_info):
        """Setup the Cloudera Management Service

        :param service_setup_info: ApiServiceSetupInfo object.
        :return: The management service instance.
        """
        return self._put("service", ApiService, data=service_setup_info)

    def get_service(self):
        """Return the Cloudera Management Services instance

        :return: An ApiService instance.
        """
        return self._get("service", ApiService)

    def hosts_start_roles(self, host_names):
        """Start all the roles on the specified hosts

        :param host_names: List of names of hosts on which to start all roles.
        :return: Information about the submitted command.
        :since: API v2
        """
        return self._cmd('hostsStartRoles', data=host_names)

    def update_config(self, config):
        """Update the CM configuration.

        :param config: Dictionary with configuration to update.
        :return: Dictionary with updated configuration.
        """
        return self._update_config("config", config)

    def import_admin_credentials(self, username, password):
        """Imports the KDC Account Manager credentials needed by Cloudera

        Manager to create kerberos principals needed by CDH services.

        :param username Username of the Account Manager. Full name including
            the Kerberos realm must be specified.
        :param password Password for the Account Manager.
        :return: Information about the submitted command.
        :since: API v7
        """
        return self._cmd('importAdminCredentials', params=dict(
            username=username, password=password))



# Copyright (c) 2014 Intel Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# The contents of this file are mainly copied from cm_api sources,
# released by Cloudera. Codes not used by Sahara CDH plugin are removed.
# You can find the original codes at
#
#     https://github.com/cloudera/cm_api/tree/master/python/src/cm_api
#
# To satisfy the pep8 and python3 tests, we did some changes to the codes.
# We also change some importings to use Sahara inherited classes.

import http.cookiejar
import posixpath
import urllib

from oslo_log import log as logging
from oslo_serialization import jsonutils as json

from sahara_plugin_cdh.plugins.cdh import exceptions as ex

LOG = logging.getLogger(__name__)


class HttpClient(object):
    """Basic HTTP client tailored for rest APIs."""
    def __init__(self, base_url, exc_class=ex.CMApiException):
        """Init Method

        :param base_url: The base url to the API.
        :param exc_class: An exception class to handle non-200 results.

        Creates an HTTP(S) client to connect to the Cloudera Manager API.
        """
        self._base_url = base_url.rstrip('/')
        self._exc_class = exc_class
        self._headers = {}

        # Make a basic auth handler that does nothing. Set credentials later.
        self._passmgr = urllib.request.HTTPPasswordMgrWithDefaultRealm()
        authhandler = urllib.request.HTTPBasicAuthHandler(self._passmgr)

        # Make a cookie processor
        cookiejar = http.cookiejar.CookieJar()

        self._opener = urllib.request.build_opener(
            urllib.request.HTTPErrorProcessor(),
            urllib.request.HTTPCookieProcessor(cookiejar),
            authhandler)

    def set_basic_auth(self, username, password, realm):
        """Set up basic auth for the client

        :param username: Login name.
        :param password: Login password.
        :param realm: The authentication realm.
        :return: The current object
        """
        self._passmgr.add_password(realm, self._base_url, username, password)
        return self

    def set_headers(self, headers):
        """Add headers to the request

        :param headers: A dictionary with the key value pairs for the headers
        :return: The current object
        """
        self._headers = headers
        return self

    @property
    def base_url(self):
        return self._base_url

    def _get_headers(self, headers):
        res = self._headers.copy()
        if headers:
            res.update(headers)
        return res

    def execute(self, http_method, path, params=None, data=None, headers=None):
        """Submit an HTTP request

        :param http_method: GET, POST, PUT, DELETE
        :param path: The path of the resource.
        :param params: Key-value parameter data.
        :param data: The data to attach to the body of the request.
        :param headers: The headers to set for this request.

        :return: The result of urllib.request.urlopen()
        """
        # Prepare URL and params
        url = self._make_url(path, params)
        if http_method in ("GET", "DELETE"):
            if data is not None:
                LOG.warning("{method} method does not pass any data. "
                            "Path {path}".format(method=http_method,
                                                 path=path))
                data = None

        if http_method in ("POST", "PUT"):
            if data is not None:
                data = data.encode('utf-8')

        # Setup the request

        request = urllib.request.Request(url, data)
        # Hack/workaround because urllib2 only does GET and POST
        request.get_method = lambda: http_method

        headers = self._get_headers(headers)
        for k, v in headers.items():
            request.add_header(k, v)

        # Call it
        LOG.debug("Method: {method}, URL: {url}".format(method=http_method,
                                                        url=url))
        try:
            return self._opener.open(request)
        except urllib.error.HTTPError as ex:
            message = str(ex)
            try:
                json_body = json.loads(message)
                message = json_body['message']
            except (ValueError, KeyError):
                pass    # Ignore json parsing error
            raise self._exc_class(message)

    def _make_url(self, path, params):
        res = self._base_url
        if path:
            res += posixpath.normpath('/' + path.lstrip('/'))
        if params:
            param_str = urllib.parse.urlencode(params, True)
            res += '?' + param_str
        return res
